import logging
import time

import pandas as pd

from carball.json_parser.game import Game
from .rotations import predict_user_inputs

# RETURNS CONTROLS
# THROTTLE STEER PITCH YAW ROLL JUMP BOOST HANDBRAKE

logger = logging.getLogger(__name__)


class ControlsCreator:

    def get_controls(self, game: Game):
        logger.info('Creating controls')
        start_time = time.time()
        for player in game.players:
            logger.debug('getting controls for player %s', player.name)
            throttle = player.data.throttle / 128 - 1
            steer = -(player.data.steer / 128 - 1)

            _jump = player.data.jump_active % 2 == 1
            _double_jump_active = player.data.double_jump_active % 2 == 1
            _dodge_active = player.data.dodge_active % 2 == 1
            jump = _jump | _double_jump_active | _dodge_active
            boost = player.data.boost_active
            handbrake = player.data.handbrake

            frames_not_on_ground = player.data.loc[:, 'pos_z'][player.data.loc[:, 'pos_z'] > 18].index.values
            # print(frames_not_on_ground)
            rotations = player.data.loc[frames_not_on_ground, ['rot_x', 'rot_y', 'rot_z']]
            ang_vels = player.data.loc[frames_not_on_ground, ['ang_vel_x', 'ang_vel_y', 'ang_vel_z']] / 1000

            predicted_inputs = predict_user_inputs(ang_vels, rotations, game.frames.delta)
            # print(predicted_inputs)
            pitch = predicted_inputs.loc[:, 'predicted_input_pitch']
            yaw = predicted_inputs.loc[:, 'predicted_input_yaw']
            roll = predicted_inputs.loc[:, 'predicted_input_roll']

            # rotations = pd.concat((player.data.pos_z, player.data.loc[frames_not_on_ground, 'rot_x':'rot_z'],
            #                        predicted_inputs), axis=1)

            player.controls = pd.DataFrame.from_dict({'throttle': throttle, 'steer': steer, 'pitch': pitch, 'yaw': yaw,
                                                      'roll': roll, 'jump': jump, 'boost': boost,
                                                      'handbrake': handbrake})

        logger.info('Finished controls in %s seconds', str(time.time() - start_time))
        return



import logging

import pandas as pd

from ...generated.api import game_pb2
from ...json_parser.game import Game


logger = logging.getLogger(__name__)


def remove_invalid_players(game: Game, data_frame: pd.DataFrame, proto_game: game_pb2.Game, player_map):
    invalid_players = []
    for player in proto_game.players:
        name = player.name
        if player.time_in_game < 5 or name not in data_frame or ('pos_x' not in data_frame[name]):
            invalid_players.append(player.name)
    if len(invalid_players) == 0:
        return

    # remove from game
    game.players = [player for player in game.players if player.name not in invalid_players]

    # remove from protobuf
    proto_players = proto_game.players
    i = 0
    total_length = len(proto_players)

    while i < total_length:
        player = proto_players[i]
        if player.name in invalid_players:
            del player_map[player.id.id]
            del proto_players[i]
            i -= 1
            total_length -= 1

            # remove player from team
            for team in proto_game.teams:
                for i in range(len(team.player_ids)):
                    if team.player_ids[i] == player.id:
                        del team.player_ids[i]
                        break
        i += 1

    logger.warning('removed player: ' + str(invalid_players))



from typing import List, Callable

from ....generated.api import team_pb2, game_pb2
from ....json_parser.game import Game


class ApiTeam:
    @staticmethod
    def create_teams_from_game(game: Game, proto_game: game_pb2.Game, id_creator: Callable) -> List[team_pb2.Team]:
        teams = []
        for team in game.teams:
            proto_team = proto_game.teams.add()
            if team.name is not None:
                proto_team.name = str(team.name)
            for player in team.players:
                id_creator(proto_team.player_ids.add(), player.name)
            proto_team.is_orange = team.is_orange
            proto_team.score = team.score
        return teams



from typing import TypeVar, Generic, List, Optional, ClassVar

from ..errors.errors import CheckError, CheckErrorLevel

T = TypeVar('T')


class BaseCheck(Generic[T]):
    error_class = CheckError
    class_message: ClassVar[str]

    def __init__(self, obj: T):
        self.obj = obj

    def run_check(self) -> List[CheckError]:
        check_functions_to_error_levels = {
            self.critical_checks: CheckErrorLevel.CRITICAL,
            self.major_checks: CheckErrorLevel.MAJOR,
            self.minor_checks: CheckErrorLevel.MINOR
        }
        errors = []
        for check_function, error_level in check_functions_to_error_levels.items():
            errors += [
                self.error_class(error_level, f"{self.class_message}: {e}")
                for e in check_function() if e is not None
            ]

        return errors

    def critical_checks(self) -> List[Optional[AssertionError]]:
        return []

    def major_checks(self) -> List[Optional[AssertionError]]:
        return []

    def minor_checks(self) -> List[Optional[AssertionError]]:
        return []

    @staticmethod
    def check(condition: bool, message: str) -> Optional[AssertionError]:
        if not condition:
            return AssertionError(message)



class ApiPlayerCameraSettings:

    @staticmethod
    def create_from_player(proto_camera, player):
        player_camera_settings = player.camera_settings

        if 'stiffness' in player.camera_settings and player.camera_settings['stiffness'] is not None:
            proto_camera.stiffness = player_camera_settings['stiffness']
        if 'height' in player.camera_settings and player.camera_settings['height'] is not None:
            proto_camera.height = player_camera_settings['height']
        if 'transition_speed' in player.camera_settings and player.camera_settings['transition_speed'] is not None:
            proto_camera.transition_speed = player_camera_settings['transition_speed']
        if 'pitch' in player.camera_settings and player.camera_settings['pitch'] is not None:
            proto_camera.pitch = player_camera_settings['pitch']
        if 'swivel_speed' in player.camera_settings and player.camera_settings['swivel_speed'] is not None:
            proto_camera.swivel_speed = player_camera_settings['swivel_speed']
        if 'field_of_view' in player.camera_settings and player.camera_settings['field_of_view'] is not None:
            proto_camera.field_of_view = player_camera_settings['field_of_view']
        if 'distance' in player.camera_settings and player.camera_settings['distance'] is not None:
            proto_camera.distance = player_camera_settings['distance']



import logging
import os

import pandas as pd

logger = logging.getLogger(__name__)


def read_xl() -> pd.DataFrame:
    car_item_dict = {
        "21": "Backfire", "22": "Breakout", "23": "Octane", "24": "Paladin", "25": "Road Hog", "26": "Gizmo",
        "28": "X-Devil", "29": "Hotshot", "30": "Merc", "31": "Venom", "402": "Takumi", "403": "Dominus",
        "404": "Scarab", "523": "Zippy", "597": "DeLorean Time Machine", "600": "Ripper", "607": "Grog",
        "803": "Batmobile", "1018": "Dominus GT", "1159": "X-Devil Mk2", "1171": "Masamune", "1172": "Marauder",
        "1286": "Aftershock", "1295": "Takumi RX-T", "1300": "Road Hog XL", "1317": "Esper", "1416": "Breakout Type-S",
        "1475": "Proteus", "1478": "Triton", "1533": "Vulcan", "1568": "Octane ZSR", "1603": "Twin Mill III",
        "1623": "Bone Shaker", "1624": "Endo", "1675": "Ice Charger", "1691": "Mantis", "1856": "J\u00e4ger 619 RS",
        "1919": "Centio V17", "1932": "Animus GP", "2268": "'70 Dodge Charger R/T",
        "2269": "'99 Nissan Skyline GT-R R34"
    }
    car_name_to_item = {value: int(key) for key, value in car_item_dict.items()}

    excel_file_name = [i for i in os.listdir(os.path.dirname(os.path.realpath(__file__))) if i.endswith('xlsx')][0]
    excel_file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), excel_file_name)
    xl = pd.read_excel(excel_file_path)

    indices = []
    for row_number, row_data in xl.iterrows():
        car = row_data['Car']
        car_index = car_name_to_item.get(car, row_number)
        indices.append(car_index)
    xl['Item ID'] = indices
    xl.set_index('Item ID', inplace=True)
    return xl



import json
import logging
import re
from datetime import datetime
from typing import List, Dict

import pandas as pd

from .actor_parsing import BallActor, CarActor
from carball.generated.api.game_pb2 import mutators_pb2 as mutators
from .goal import Goal
from .player import Player
from .team import Team
from .game_info import GameInfo

logger = logging.getLogger(__name__)

BOOST_PER_SECOND = 80 * 1 / .93  # boost used per second out of 255
DATETIME_FORMATS = [
    '%Y-%m-%d %H-%M-%S',
    '%Y-%m-%d:%H-%M'
]


class Game:

    def __init__(self):
        self.file_path = None
        self.replay = None
        self.properties = None
        self.replay_id = None
        self.map = None
        self.name = None
        self.id = None
        self.datetime = None
        self.replay_version = None

        self.replay_data = None

        self.teams: List[Team] = None  # Added in parse_all_data
        self.players: List[Player] = None
        self.goals: List[Goal] = None
        self.primary_player: dict = None
        self.all_data = None

        self.frames = None
        self.kickoff_frames = None
        self.ball = None
        self.ball_type = None
        self.demos = None
        self.parties = None

    def initialize(self, file_path='', loaded_json=None, parse_replay: bool = True, clean_player_names: bool = False):
        self.file_path = file_path
        if loaded_json is None:
            with open(file_path, 'r') as f:
                self.replay = json.load(f)
        else:
            self.replay = loaded_json
        logger.debug('Loaded JSON')

        self.replay_data = self.replay['content']['body']['frames']

        # set properties
        self.properties = self.replay['header']['body']['properties']['value']
        self.replay_id = self.find_actual_value(self.properties['Id']['value'])
        if 'MapName' in self.properties:
            self.map = self.find_actual_value(self.properties['MapName']['value'])
        else:
            self.map = 'Unknown'
        self.name = self.find_actual_value(self.properties.get('ReplayName', None))
        self.match_type = self.find_actual_value(self.properties['MatchType']['value'])
        self.team_size = self.find_actual_value(self.properties['TeamSize']['value'])

        if self.name is None:
            logger.warning('Replay name not found')
        self.id = self.find_actual_value(self.properties["Id"]['value'])

        date_string = self.properties['Date']['value']['str']
        for date_format in DATETIME_FORMATS:
            try:
                self.datetime = datetime.strptime(date_string, date_format)
                break
            except ValueError:
                pass
        else:
            logger.error('Cannot parse date: ' + date_string)

        self.replay_version = self.properties.get('ReplayVersion', {}).get('value', {}).get('int', None)
        logger.info(f"version: {self.replay_version}, date: {self.datetime}")
        if self.replay_version is None:
            logger.warning('Replay version not found')

        self.players: List[Player] = self.create_players()
        self.goals: List[Goal] = self.get_goals()
        self.primary_player: dict = self.get_primary_player()

        if parse_replay:
            self.all_data = self.parse_replay()
            self.parse_all_data(self.all_data, clean_player_names)
            logger.info("Finished parsing %s" % self)

    def __repr__(self):
        team_0_name = self.teams[0].name
        team_1_name = self.teams[1].name
        team_0_score = self.teams[0].score
        team_1_score = self.teams[1].score
        return "%s: %s vs %s (%s:%s)" % (self.name, team_0_name, team_1_name, team_0_score, team_1_score)

    def create_players(self) -> List[Player]:
        players = []
        try:
            for player_stats in self.properties["PlayerStats"]["value"]["array"]:
                player = Player().parse_player_stats(player_stats["value"])
                players.append(player)
        except KeyError:
            pass
        return players

    def get_primary_player(self):
        owner_name = self.properties.get('PlayerName')
        if owner_name is not None:
            owner_name = owner_name['value']['str']
            for player in self.players:
                if player.name == owner_name:
                    return {'name': owner_name, 'id': player.online_id}
            return {'name': owner_name, 'id': None}

    def get_goals(self) -> List[Goal]:
        if "Goals" not in self.properties:
            return []

        goals = [g['value'] for g in self.properties["Goals"]["value"]["array"]]

        logger.info('Found %s goals.' % len(goals))
        logger.debug('Goals: %s' % goals)

        goals_list = []
        for goal_dict in goals:
            goal = Goal(goal_dict, self)
            goals_list.append(goal)
        return goals_list

    @staticmethod
    def find_actual_value(value_dict: dict) -> dict or int or bool or str:
        types = ['int', 'boolean', 'string', 'byte', 'str', 'name', ('flagged_int', 'int')]
        if value_dict is None:
            return None
        if 'value' in value_dict:
            value_dict = value_dict['value']
        for _type in types:
            if isinstance(_type, str):
                if _type in value_dict:
                    return value_dict[_type]
            else:
                value = value_dict
                if _type[0] in value:
                    for type_str in _type:
                        value = value[type_str]
                    return value
        else:
            return value_dict

    def parse_replay(self):
        """
        :return: all_data = {
            'player_ball_data': player_ball_data,
            'player_dicts': player_dicts,
            'team_dicts': team_dicts,
            'frames_data': frames_data,
            'cameras_data': cameras_data
            'demos_data': demos_data
        }

        player_ball_data format:
        {
        'ball': {frame_number: {pos_x, pos_y ...}, f_no2: {...} ...,}
        player_actor_id: {
            frame_number: {
                pos_x, pos_y ...,
                throttle, steer, ...,
                ping, ball_cam
            },
            f_no2: {...} ...,
        }

        currently implemented:
            inputs: posx, posy, posz, rotx, roty, rotz, vx, vy, vz, angvx, angy, angvz, boost_amt
            outputs: throttle, steer, handbrake, boost, jump, doublejump, dodge

        player_dicts  = {player_actor_id : {actor_data}, player_actor_id_2: {actor_data_2}}
        team_dicts = {team_actor_id: {actor_data, 'colour':'blue'/'orange', also includes name}
        frames_data = {frame_number: {time, delta, seconds_remaining, is_overtime, ball_has_been_hit}
        cameras_data = {player_actor_id: actor_data}
        demos_data = {frame_number: demolish_data}

        """
        player_ball_data = {'ball': {}}

        # dictionaries to contain data in frames
        player_dicts = {}  # player_actor_id: player_dict
        team_dicts = {}

        player_car_ids = {}  # player_actor_id: car_actor_id
        car_player_ids = {}  # car_actor_id: player_actor_id

        current_actor_ids = []
        current_actors = {}  # id:actor_update
        # stores car_actor_ids to collect data for at each frame
        current_car_ids_to_collect = []

        frames = self.replay_data
        frames_data = {}
        soccar_game_event_actor_id = None
        frame_number = 0
        current_goal_number = 0

        cameras_data = {}  # player_actor_id: actor_data
        demos_data = []  # frame_number: demolish_data

        game_info_actor = None
        parties = {}
        # loop through frames
        for i, frame in enumerate(frames):
            # # don't bother after last goal
            # if frame_number > last_goal_frame_number:
            #     break

            _f_time = frame["time"]
            _f_delta = frame["delta"]
            deleted_actor_ids = []
            actor_updates = []
            actor_spawns = []
            for replication in frame['replications']:
                category = list(replication['value'].keys())[0]
                if category == 'destroyed':
                    deleted_actor_ids.append(replication)
                elif category == 'updated':
                    actor_updates.append(replication)
                elif category == 'spawned':
                    actor_spawns.append(replication)
            # remove deleted actors
            for deleted_actor in deleted_actor_ids:
                deleted_actor_id = deleted_actor['actor_id']['value']
                current_actor_ids.remove(deleted_actor_id)
                current_actors.pop(deleted_actor_id)
                if deleted_actor_id in car_player_ids:
                    player_actor_id = car_player_ids[deleted_actor_id]
                    car_player_ids.pop(deleted_actor_id)
                    try:
                        player_car_ids.pop(player_actor_id)
                    except KeyError:
                        pass

            for actor_spawn in actor_spawns:
                actor_id = actor_spawn['actor_id']['value']
                current_actor_ids.append(actor_id)
                current_actors[actor_id] = {
                    'TypeName': actor_spawn['value']['spawned']['object_name'],
                    'ClassName': actor_spawn['value']['spawned']['class_name'],
                    'Name': actor_spawn['value']['spawned'].get('name', None),
                    'Id': actor_id
                }
            # apply actor updates
            for actor_update in actor_updates:
                actor_id = actor_update['actor_id']['value']
                # update_type = list(actor_update['value'].keys())[0]
                actual_update = {v['name']: self.find_actual_value(v['value']) for v in
                                 actor_update['value']['updated']}
                # TODO: process each subtype to a single value
                # add if new actor
                if actor_id not in current_actor_ids:
                    current_actor_ids.append(actor_id)
                    current_actors[actor_id] = actual_update
                else:
                    # update stuff in current_actors
                    for _k, _v in actual_update.items():
                        current_actors[actor_id][_k] = _v

            # server metadata
            if game_info_actor is None:
                for actor_id, actor_data in current_actors.items():
                    if actor_data['TypeName'].endswith(':GameReplicationInfoArchetype'):
                        game_info_actor = actor_data
                        break

            # find players and ball
            players_and_teams_data = self.parse_players_and_teams(current_actors, parties, player_dicts,
                                                                  player_ball_data, team_dicts)
            parties = players_and_teams_data["parties"]
            player_dicts = players_and_teams_data["player_dicts"]
            player_ball_data = players_and_teams_data["player_ball_data"]
            team_dicts = players_and_teams_data["team_dicts"]

            # stop data collection after goal
            REPLICATED_RB_STATE_KEY = "TAGame.RBActor_TA:ReplicatedRBState"
            try:
                if frame_number > self.goals[current_goal_number].frame_number:
                    # set all players to sleeping after goal
                    for car_actor_id in car_player_ids:
                        try:
                            car = current_actors[car_actor_id]
                            car[REPLICATED_RB_STATE_KEY]['Sleeping'] = True
                        except KeyError as e:
                            # Ignore the case where the car does not have a REPLICATED_RB_STATE_KEY
                            pass
                    current_goal_number += 1
            except IndexError:
                # after last goal.
                pass

            # gather data at this frame
            if _f_delta != 0:
                # frame stuff
                frame_data = {
                    'time': frame["time"],
                    'delta': frame["delta"],
                }
                if soccar_game_event_actor_id is None:
                    # set soccar_game_event_actor_id
                    for actor_id, actor_data in current_actors.items():
                        if actor_data["TypeName"] == "Archetypes.GameEvent.GameEvent_Soccar" \
                                or "TAGame.GameEvent_Soccar_TA:SecondsRemaining" in actor_data:
                            # TODO: Investigate if there's a less hacky way to detect GameActors with not TypeName
                            soccar_game_event_actor_id = actor_id
                            break

                frame_data['seconds_remaining'] = current_actors[soccar_game_event_actor_id].get(
                    "TAGame.GameEvent_Soccar_TA:SecondsRemaining", None)
                frame_data['replicated_seconds_remaining'] = current_actors[soccar_game_event_actor_id].get(
                    "TAGame.GameEvent_TA:ReplicatedGameStateTimeRemaining", None)
                frame_data['is_overtime'] = current_actors[soccar_game_event_actor_id].get(
                    "TAGame.GameEvent_Soccar_TA:bOverTime", None)
                frame_data['ball_has_been_hit'] = current_actors[soccar_game_event_actor_id].get(
                    "TAGame.GameEvent_Soccar_TA:bBallHasBeenHit", None)
                frames_data[frame_number] = frame_data

                # car and player stuff
                for actor_id, actor_data in current_actors.items():
                    if actor_data["TypeName"] == "Archetypes.Car.Car_Default" and \
                            "Engine.Pawn:PlayerReplicationInfo" in actor_data:
                        player_actor_id = actor_data[
                            "Engine.Pawn:PlayerReplicationInfo"]
                        # assign car player links
                        player_car_ids[player_actor_id] = actor_id
                        car_player_ids[actor_id] = player_actor_id

                        RBState = actor_data.get(
                            REPLICATED_RB_STATE_KEY, {})
                        # bDriving is missing?! TODO: Investigate bDriving in RBState
                        # car_is_driving = RBState.get("rigid_body_state", {}).get("TAGame.Vehicle_TA:bDriving", False)
                        car_is_sleeping = RBState.get("rigid_body_state", {}).get('sleeping', True)
                        # only collect data if car is driving and not sleeping
                        if not car_is_sleeping:
                            current_car_ids_to_collect.append(actor_id)

                            data_dict = CarActor.get_data_dict(actor_data, version=self.replay_version)
                            # save data from here
                            player_ball_data[player_actor_id][frame_number] = data_dict

                        # get demo data
                        if "TAGame.Car_TA:ReplicatedDemolish" in actor_data:

                            demo_data = actor_data["TAGame.Car_TA:ReplicatedDemolish"]['demolish']
                            # add attacker and victim player ids
                            attacker_car_id = demo_data["attacker_actor_id"]
                            victim_car_id = demo_data["victim_actor_id"]
                            if attacker_car_id != -1 and victim_car_id != -1 and \
                                    attacker_car_id < 1e9 and victim_car_id < 1e9:
                                # Filter out weird stuff where it's not a demo
                                # frame 1 of 0732D41D4AF83D610AE2A988ACBC977A (rlcs season 4 eu)
                                attacker_player_id = car_player_ids[attacker_car_id]
                                victim_player_id = car_player_ids[victim_car_id]
                                demo_data["attacker_player_id"] = attacker_player_id
                                demo_data["victim_player_id"] = victim_player_id
                                # add frame_number
                                demo_data['frame_number'] = frame_number
                                demos_data.append(demo_data)
                                logger.debug('ReplicatedDemolish: Att: %s, Def: %s' %
                                             (attacker_player_id, victim_player_id))
                                logger.debug('RepDemo Names: Att: %s. Def: %s' %
                                             (player_dicts[attacker_player_id]['name'],
                                              player_dicts[victim_player_id]['name']))
                                actor_data.pop("TAGame.Car_TA:ReplicatedDemolish")

                    elif actor_data["TypeName"] == "Archetypes.Ball.Ball_Default":
                        if actor_data.get('TAGame.RBActor_TA:bIgnoreSyncing', False):
                            continue
                        self.ball_type = mutators.DEFAULT
                        # RBState = actor_data.get(REPLICATED_RB_STATE_KEY, {})
                        # ball_is_sleeping = RBState.get('Sleeping', True)
                        data_dict = BallActor.get_data_dict(actor_data, version=self.replay_version)
                        player_ball_data['ball'][frame_number] = data_dict
                    elif actor_data["TypeName"] == "Archetypes.Ball.Ball_Basketball" or \
                            actor_data["TypeName"] == "Archetypes.Ball.Ball_BasketBall":
                        if actor_data.get('TAGame.RBActor_TA:bIgnoreSyncing', False):
                            continue
                        self.ball_type = mutators.BASKETBALL
                        data_dict = BallActor.get_data_dict(actor_data, version=self.replay_version)
                        player_ball_data['ball'][frame_number] = data_dict
                    elif actor_data["TypeName"] == "Archetypes.Ball.Ball_Puck":
                        if actor_data.get('TAGame.RBActor_TA:bIgnoreSyncing', False):
                            continue
                        self.ball_type = mutators.PUCK
                        data_dict = BallActor.get_data_dict(actor_data, version=self.replay_version)
                        player_ball_data['ball'][frame_number] = data_dict
                    elif actor_data["TypeName"] == "Archetypes.Ball.CubeBall":
                        if actor_data.get('TAGame.RBActor_TA:bIgnoreSyncing', False):
                            continue
                        self.ball_type = mutators.CUBEBALL
                        data_dict = BallActor.get_data_dict(actor_data, version=self.replay_version)
                        player_ball_data['ball'][frame_number] = data_dict
                    elif actor_data["TypeName"] == "Archetypes.Ball.Ball_Breakout":
                        if actor_data.get('TAGame.RBActor_TA:bIgnoreSyncing', False):
                            continue
                        self.ball_type = mutators.BREAKOUT
                        data_dict = BallActor.get_data_dict(actor_data, version=self.replay_version)
                        player_ball_data['ball'][frame_number] = data_dict

                for actor_id, actor_data in current_actors.items():
                    COMPONENT_ACTIVE_KEY = "TAGame.CarComponent_TA:Active"
                    COMPONENT_REPLICATED_ACTIVE_KEY = "TAGame.CarComponent_TA:ReplicatedActive"
                    if actor_data["TypeName"] == "TAGame.Default__CameraSettingsActor_TA" and \
                            "TAGame.CameraSettingsActor_TA:PRI" in actor_data:
                        player_actor_id = actor_data["TAGame.CameraSettingsActor_TA:PRI"]  # may need to try another key
                        # add camera settings
                        if player_actor_id not in cameras_data and \
                                "TAGame.CameraSettingsActor_TA:ProfileSettings" in actor_data:
                            cameras_data[player_actor_id] = actor_data["TAGame.CameraSettingsActor_TA:ProfileSettings"]
                        # add ball cam to inputs
                        ball_cam = actor_data.get("TAGame.CameraSettingsActor_TA:bUsingSecondaryCamera", None)
                        try:
                            player_ball_data[player_actor_id][frame_number]['ball_cam'] = ball_cam
                        except KeyError:
                            # key error due to frame_number not in inputs
                            # ignore as no point knowing
                            pass
                    elif actor_data["TypeName"] == "TAGame.Default__PRI_TA" and \
                            "TAGame.PRI_TA:CameraSettings" in actor_data:
                        # oldstyle camera settings
                        if actor_id not in cameras_data:
                            cameras_data[actor_id] = actor_data["TAGame.PRI_TA:CameraSettings"]
                        ball_cam = actor_data.get("TAGame.CameraSettingsActor_TA:bUsingSecondaryCamera", None)
                        try:
                            player_ball_data[actor_id][frame_number]['ball_cam'] = ball_cam
                        except KeyError:
                            # key error due to frame_number not in inputs
                            # ignore as no point knowing
                            pass
                    elif actor_data["TypeName"] == "Archetypes.CarComponents.CarComponent_Boost":
                        car_actor_id = actor_data.get(
                            "TAGame.CarComponent_TA:Vehicle", None)
                        if car_actor_id is not None:
                            if car_actor_id in current_car_ids_to_collect:
                                player_actor_id = car_player_ids[car_actor_id]
                                boost_is_active_random_int = actor_data.get(
                                    COMPONENT_ACTIVE_KEY,
                                    actor_data.get(COMPONENT_REPLICATED_ACTIVE_KEY, False))
                                # boost_is_active when random_int is odd?!
                                boost_is_active = (boost_is_active_random_int % 2 == 1)
                                if boost_is_active:
                                    # manually decrease car boost amount (not shown in replay)
                                    # i assume game calculates the decrease itself similarly
                                    boost_amount = max(0,
                                                       actor_data.get(
                                                           "TAGame.CarComponent_Boost_TA:ReplicatedBoostAmount", 0) -
                                                       _f_delta * BOOST_PER_SECOND)
                                    actor_data["TAGame.CarComponent_Boost_TA:ReplicatedBoostAmount"] = boost_amount
                                else:
                                    boost_amount = actor_data.get("TAGame.CarComponent_Boost_TA:ReplicatedBoostAmount",
                                                                  None)

                                player_ball_data[player_actor_id][frame_number]['boost'] = boost_amount
                                player_ball_data[player_actor_id][frame_number]['boost_active'] = boost_is_active
                    elif actor_data["TypeName"] == "Archetypes.CarComponents.CarComponent_Jump":
                        car_actor_id = actor_data.get(
                            "TAGame.CarComponent_TA:Vehicle", None)
                        if car_actor_id is not None:
                            if car_actor_id in current_car_ids_to_collect:
                                player_actor_id = car_player_ids[car_actor_id]
                                jump_is_active = actor_data.get(
                                    COMPONENT_ACTIVE_KEY,
                                    actor_data.get(COMPONENT_REPLICATED_ACTIVE_KEY, False))
                                player_ball_data[player_actor_id][frame_number]['jump_active'] = jump_is_active
                    elif actor_data["TypeName"] == "Archetypes.CarComponents.CarComponent_DoubleJump":
                        car_actor_id = actor_data.get(
                            "TAGame.CarComponent_TA:Vehicle", None)
                        if car_actor_id is not None:
                            if car_actor_id in current_car_ids_to_collect:
                                player_actor_id = car_player_ids[car_actor_id]
                                double_jump_is_active = actor_data.get(
                                    COMPONENT_ACTIVE_KEY,
                                    actor_data.get(COMPONENT_REPLICATED_ACTIVE_KEY, False))
                                player_ball_data[player_actor_id][frame_number][
                                    'double_jump_active'] = double_jump_is_active
                    elif actor_data["TypeName"] == "Archetypes.CarComponents.CarComponent_Dodge":
                        car_actor_id = actor_data.get(
                            "TAGame.CarComponent_TA:Vehicle", None)
                        if car_actor_id is not None:
                            if car_actor_id in current_car_ids_to_collect:
                                player_actor_id = car_player_ids[car_actor_id]
                                dodge_is_active = actor_data.get(
                                    COMPONENT_ACTIVE_KEY,
                                    actor_data.get(COMPONENT_REPLICATED_ACTIVE_KEY, False))
                                player_ball_data[player_actor_id][frame_number]['dodge_active'] = dodge_is_active
                    elif actor_data["ClassName"] == "TAGame.VehiclePickup_Boost_TA":
                        REPLICATED_PICKUP_KEY = "TAGame.VehiclePickup_TA:ReplicatedPickupData"
                        if REPLICATED_PICKUP_KEY in actor_data and \
                                actor_data[REPLICATED_PICKUP_KEY] != -1 and \
                                'instigator_id' in actor_data[REPLICATED_PICKUP_KEY]['pickup']:
                            car_actor_id = actor_data[REPLICATED_PICKUP_KEY]['pickup']['instigator_id']
                            if car_actor_id in car_player_ids:
                                player_actor_id = car_player_ids[car_actor_id]
                                if frame_number in player_ball_data[player_actor_id]:
                                    player_ball_data[player_actor_id][frame_number]['boost_collect'] = True
                                    # TODO: Investigate and fix random imaginary boost collects
                                # set to false after acknowledging it's turned True
                                # it does not turn back false immediately although boost is only collected once.
                                # using actor_id!=-1
                                actor_data[REPLICATED_PICKUP_KEY]['pickup']["instigator_id"] = -1
                for player_actor_id in player_dicts:
                    actor_data = current_actors.get(player_actor_id, None)
                    if actor_data is None:
                        continue
                    if frame_number in player_ball_data[player_actor_id]:
                        player_ball_data[player_actor_id][frame_number]['ping'] = actor_data.get(
                            "Engine.PlayerReplicationInfo:Ping", None)
                    else:
                        player_ball_data[player_actor_id][frame_number] = {
                            'ping': actor_data.get("Engine.PlayerReplicationInfo:Ping", None)}

            current_car_ids_to_collect = []  # reset ids for next frame
            frame_number += 1

        all_data = {
            'player_ball_data': player_ball_data,
            'player_dicts': player_dicts,
            'team_dicts': team_dicts,
            'frames_data': frames_data,
            'cameras_data': cameras_data,
            'demos_data': demos_data,
            'game_info_actor': game_info_actor,
            'parties': parties
        }

        return all_data

    def parse_players_and_teams(self, current_actors, parties, player_dicts, player_ball_data, team_dicts) -> Dict:
        """
              :return: players_and_team_data = {
                  'parties': parties,
                  'player_dicts': player_dicts,
                  'player_ball_data': player_ball_data,
                  'team_dicts': team_dicts,
              }
        """

        for actor_id, actor_data in current_actors.items():
            if actor_data["TypeName"] == "TAGame.Default__PRI_TA" \
                    and "Engine.PlayerReplicationInfo:PlayerName" in actor_data:

                player_dict = {
                    'name': actor_data["Engine.PlayerReplicationInfo:PlayerName"],
                }
                # Conditionally add ['team'] key to player_dict
                player_team = actor_data.get("Engine.PlayerReplicationInfo:Team", None)
                if player_team is not None and player_team != -1:
                    player_dict['team'] = player_team

                if "TAGame.PRI_TA:PartyLeader" in actor_data:
                    try:
                        actor_type = \
                            list(actor_data["Engine.PlayerReplicationInfo:UniqueId"]['unique_id'][
                                     'remote_id'].keys())[
                                0]

                        # handle UniqueID for plays_station and switch
                        if actor_type == "play_station" or actor_type == "psy_net":
                            actor_name = actor_data["Engine.PlayerReplicationInfo:PlayerName"]
                            for player_stat in self.properties['PlayerStats']['value']["array"]:
                                if actor_name == player_stat['value']['Name']['value']['str']:
                                    unique_id = str(player_stat['value']['OnlineID']['value']['q_word'])
                        else:
                            unique_id = str(
                                actor_data['Engine.PlayerReplicationInfo:UniqueId']['unique_id']['remote_id'][actor_type])

                        # only process if party_leader id exists
                        if "party_leader" in actor_data["TAGame.PRI_TA:PartyLeader"] and \
                                "id" in actor_data["TAGame.PRI_TA:PartyLeader"]["party_leader"]:
                            leader_actor_type = list(
                                actor_data["TAGame.PRI_TA:PartyLeader"]["party_leader"]["id"][0].keys()
                            )[0]
                            if leader_actor_type == "play_station" or leader_actor_type == "psy_net":
                                leader_name = actor_data[
                                    "TAGame.PRI_TA:PartyLeader"
                                ]["party_leader"]["id"][0][leader_actor_type][0]

                                for player_stat in self.properties['PlayerStats']['value']["array"]:
                                    if leader_name == player_stat['value']['Name']['value']['str']:
                                        leader = str(player_stat['value']['OnlineID']['value']['q_word'])

                            else:  # leader is not using play_station nor switch (ie. xbox or steam)
                                leader = str(
                                    actor_data[
                                        "TAGame.PRI_TA:PartyLeader"
                                    ]["party_leader"]["id"][0][leader_actor_type]
                                )

                            if leader in parties:
                                if unique_id not in parties[leader]:
                                    parties[leader].append(unique_id)
                            else:
                                parties[leader] = [unique_id]

                    except KeyError:
                        logger.warning('Could not get party leader for actor id: ' + str(actor_id))

                if actor_id not in player_dicts:
                    # add new player
                    player_dicts[actor_id] = player_dict

                    logger.debug('Found player actor: %s (id: %s)' % (player_dict['name'], actor_id))
                    player_ball_data[actor_id] = {}
                else:
                    # update player_dicts
                    for _k, _v in {**actor_data, **player_dict}.items():
                        player_dicts[actor_id][_k] = _v
            elif actor_data["ClassName"] == "TAGame.Team_Soccar_TA":
                team_dicts[actor_id] = actor_data
                team_dicts[actor_id]['colour'] = 'blue' if actor_data[
                                                               "TypeName"] == "Archetypes.Teams.Team0" else 'orange'

        players_and_team_data = {
            'parties': parties,
            'player_dicts': player_dicts,
            'player_ball_data': player_ball_data,
            'team_dicts': team_dicts
        }

        return players_and_team_data

    def parse_all_data(self, all_data, clean_player_names: bool) -> None:
        """
        Finishes parsing after frame-parsing is done.
        E.g. Adds players not found in MatchStats metadata
        :param all_data: Dict returned by parse_replay
        :return:
        """
        # GAME INFO
        self.game_info = GameInfo().parse_game_info_actor(all_data['game_info_actor'])

        # TEAMS
        self.teams = []
        for team_id, team_data in all_data['team_dicts'].items():
            team = Team().parse_team_data(team_data)
            self.teams.append(team)

        # PLAYERS
        player_actor_id_player_dict = {}  # player_actor_id: Player
        for _player_actor_id, _player_data in all_data['player_dicts'].items():
            if "TAGame.PRI_TA:MatchScore" not in _player_data:
                logger.warning(f"Player {_player_data['name']} as player has no MatchScore.")

            found_player = None
            for player in self.players:
                # if player leaves early, won't be created (as not found in metadata's player_stats)
                if _player_data['name'] == player.name:
                    found_player = player
                    player_actor_id_player_dict[_player_actor_id] = found_player
                    found_player.parse_actor_data(_player_data)  # just add extra stuff
                    break
            if found_player is None:
                # player not in endgame stats, create new player
                try:
                    found_player = Player().create_from_actor_data(_player_data, self.teams)
                except KeyError as e:
                    # KeyError: 'Engine.PlayerReplicationInfo:Team'
                    # in `team_actor_id = actor_data["Engine.PlayerReplicationInfo:Team"]`
                    # Player never actually joins the game.
                    if 'Engine.PlayerReplicationInfo:Team' not in _player_data:
                        logger.warning(f"Ignoring player: {_player_data['name']} as player has no team.")
                        continue
                    raise e
                self.players.append(found_player)
                player_actor_id_player_dict[_player_actor_id] = found_player

                # check if any goals are playerless and belong to this newly-created player
                for goal in self.goals:
                    if not goal.player and goal.player_name == found_player.name:
                        goal.player = found_player

            found_player.parse_data(all_data['player_ball_data'][_player_actor_id])
            # camera_settings might not exist (see 0AF8AC734890E6D3995B829E474F9924)
            found_player.get_camera_settings(all_data['cameras_data'].get(_player_actor_id, {}).get('cam_settings', {}))

            for team in self.teams:
                if found_player.is_orange == team.is_orange:
                    team.add_player(found_player)

            if clean_player_names:
                cleaned_player_name = re.sub(r'[^\x00-\x7f]', r'', found_player.name).strip()  # Support ASCII only
                if cleaned_player_name != found_player.name:
                    logger.warning(f"Cleaned player name to ASCII-only. From: {found_player.name} to: {cleaned_player_name}")
                    found_player.name = cleaned_player_name

        # GOAL - add player if not found earlier (ie player just created)
        for goal in self.goals:
            if goal.player is None:
                goal.player = goal.get_player(self)

        # BALL
        self.ball = pd.DataFrame.from_dict(self.all_data['player_ball_data']['ball'], orient='index')

        # FRAMES
        self.frames = pd.DataFrame.from_dict(self.all_data['frames_data'], orient='index')

        # DEMOS
        self.demos = []
        for _demo_data in all_data['demos_data']:
            demo = {
                'frame_number': _demo_data['frame_number'],
                'attacker': player_actor_id_player_dict[_demo_data['attacker_player_id']],
                'victim': player_actor_id_player_dict[_demo_data['victim_player_id']],
                'attacker_vel': (_demo_data["attacker_velocity"]["x"],
                                 _demo_data["attacker_velocity"]["y"],
                                 _demo_data["attacker_velocity"]["z"],),
                'victim_vel': (_demo_data["victim_velocity"]["x"],
                               _demo_data["victim_velocity"]["y"],
                               _demo_data["victim_velocity"]["z"],),
            }
            self.demos.append(demo)

        # PARTIES
        self.parties = all_data['parties']

        del self.replay_data
        del self.replay
        del self.all_data



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/stats.proto',
  package='api.stats',
  serialized_pb=_b('\n\x15\x61pi/stats/stats.proto\x12\tapi.stats\"\x8e\x01\n\nPossession\x12\x17\n\x0fpossession_time\x18\x01 \x01(\x01\x12\x11\n\tturnovers\x18\x02 \x01(\x05\x12\x1c\n\x14turnovers_on_my_half\x18\x03 \x01(\x05\x12\x1f\n\x17turnovers_on_their_half\x18\x04 \x01(\x05\x12\x15\n\rwon_turnovers\x18\x05 \x01(\x05\"\xe8\x02\n\x14PositionalTendencies\x12\x16\n\x0etime_on_ground\x18\x01 \x01(\x02\x12\x17\n\x0ftime_low_in_air\x18\x02 \x01(\x02\x12\x18\n\x10time_high_in_air\x18\x03 \x01(\x02\x12\x1e\n\x16time_in_defending_half\x18\x04 \x01(\x02\x12\x1e\n\x16time_in_attacking_half\x18\x05 \x01(\x02\x12\x1f\n\x17time_in_defending_third\x18\x06 \x01(\x02\x12\x1d\n\x15time_in_neutral_third\x18\x07 \x01(\x02\x12\x1f\n\x17time_in_attacking_third\x18\x08 \x01(\x02\x12\x18\n\x10time_behind_ball\x18\t \x01(\x02\x12\x1a\n\x12time_in_front_ball\x18\n \x01(\x02\x12\x16\n\x0etime_near_wall\x18\x0b \x01(\x02\x12\x16\n\x0etime_in_corner\x18\x0c \x01(\x02\"e\n\x08\x41verages\x12\x15\n\raverage_speed\x18\x01 \x01(\x02\x12\x1c\n\x14\x61verage_hit_distance\x18\x02 \x01(\x02\x12$\n\x1c\x61verage_distance_from_center\x18\x03 \x01(\x02\"\xc0\x01\n\tHitCounts\x12\x12\n\ntotal_hits\x18\x01 \x01(\x05\x12\x13\n\x0btotal_goals\x18\x02 \x01(\x05\x12\x14\n\x0ctotal_passes\x18\x03 \x01(\x05\x12\x13\n\x0btotal_saves\x18\x04 \x01(\x05\x12\x13\n\x0btotal_shots\x18\x05 \x01(\x05\x12\x16\n\x0etotal_dribbles\x18\x06 \x01(\x05\x12\x1b\n\x13total_dribble_conts\x18\x07 \x01(\x05\x12\x15\n\rtotal_aerials\x18\x08 \x01(\x05\"]\n\x05Speed\x12\x1a\n\x12time_at_slow_speed\x18\x01 \x01(\x02\x12\x1b\n\x13time_at_super_sonic\x18\x02 \x01(\x02\x12\x1b\n\x13time_at_boost_speed\x18\x03 \x01(\x02')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_POSSESSION = _descriptor.Descriptor(
  name='Possession',
  full_name='api.stats.Possession',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='possession_time', full_name='api.stats.Possession.possession_time', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='turnovers', full_name='api.stats.Possession.turnovers', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='turnovers_on_my_half', full_name='api.stats.Possession.turnovers_on_my_half', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='turnovers_on_their_half', full_name='api.stats.Possession.turnovers_on_their_half', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='won_turnovers', full_name='api.stats.Possession.won_turnovers', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=37,
  serialized_end=179,
)


_POSITIONALTENDENCIES = _descriptor.Descriptor(
  name='PositionalTendencies',
  full_name='api.stats.PositionalTendencies',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='time_on_ground', full_name='api.stats.PositionalTendencies.time_on_ground', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_low_in_air', full_name='api.stats.PositionalTendencies.time_low_in_air', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_high_in_air', full_name='api.stats.PositionalTendencies.time_high_in_air', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_in_defending_half', full_name='api.stats.PositionalTendencies.time_in_defending_half', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_in_attacking_half', full_name='api.stats.PositionalTendencies.time_in_attacking_half', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_in_defending_third', full_name='api.stats.PositionalTendencies.time_in_defending_third', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_in_neutral_third', full_name='api.stats.PositionalTendencies.time_in_neutral_third', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_in_attacking_third', full_name='api.stats.PositionalTendencies.time_in_attacking_third', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_behind_ball', full_name='api.stats.PositionalTendencies.time_behind_ball', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_in_front_ball', full_name='api.stats.PositionalTendencies.time_in_front_ball', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_near_wall', full_name='api.stats.PositionalTendencies.time_near_wall', index=10,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_in_corner', full_name='api.stats.PositionalTendencies.time_in_corner', index=11,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=182,
  serialized_end=542,
)


_AVERAGES = _descriptor.Descriptor(
  name='Averages',
  full_name='api.stats.Averages',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='average_speed', full_name='api.stats.Averages.average_speed', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_hit_distance', full_name='api.stats.Averages.average_hit_distance', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_distance_from_center', full_name='api.stats.Averages.average_distance_from_center', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=544,
  serialized_end=645,
)


_HITCOUNTS = _descriptor.Descriptor(
  name='HitCounts',
  full_name='api.stats.HitCounts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='total_hits', full_name='api.stats.HitCounts.total_hits', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_goals', full_name='api.stats.HitCounts.total_goals', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_passes', full_name='api.stats.HitCounts.total_passes', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_saves', full_name='api.stats.HitCounts.total_saves', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_shots', full_name='api.stats.HitCounts.total_shots', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_dribbles', full_name='api.stats.HitCounts.total_dribbles', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_dribble_conts', full_name='api.stats.HitCounts.total_dribble_conts', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_aerials', full_name='api.stats.HitCounts.total_aerials', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=648,
  serialized_end=840,
)


_SPEED = _descriptor.Descriptor(
  name='Speed',
  full_name='api.stats.Speed',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='time_at_slow_speed', full_name='api.stats.Speed.time_at_slow_speed', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_at_super_sonic', full_name='api.stats.Speed.time_at_super_sonic', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_at_boost_speed', full_name='api.stats.Speed.time_at_boost_speed', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=842,
  serialized_end=935,
)

DESCRIPTOR.message_types_by_name['Possession'] = _POSSESSION
DESCRIPTOR.message_types_by_name['PositionalTendencies'] = _POSITIONALTENDENCIES
DESCRIPTOR.message_types_by_name['Averages'] = _AVERAGES
DESCRIPTOR.message_types_by_name['HitCounts'] = _HITCOUNTS
DESCRIPTOR.message_types_by_name['Speed'] = _SPEED

Possession = _reflection.GeneratedProtocolMessageType('Possession', (_message.Message,), dict(
  DESCRIPTOR = _POSSESSION,
  __module__ = 'api.stats.stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Possession)
  ))
_sym_db.RegisterMessage(Possession)

PositionalTendencies = _reflection.GeneratedProtocolMessageType('PositionalTendencies', (_message.Message,), dict(
  DESCRIPTOR = _POSITIONALTENDENCIES,
  __module__ = 'api.stats.stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.PositionalTendencies)
  ))
_sym_db.RegisterMessage(PositionalTendencies)

Averages = _reflection.GeneratedProtocolMessageType('Averages', (_message.Message,), dict(
  DESCRIPTOR = _AVERAGES,
  __module__ = 'api.stats.stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Averages)
  ))
_sym_db.RegisterMessage(Averages)

HitCounts = _reflection.GeneratedProtocolMessageType('HitCounts', (_message.Message,), dict(
  DESCRIPTOR = _HITCOUNTS,
  __module__ = 'api.stats.stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.HitCounts)
  ))
_sym_db.RegisterMessage(HitCounts)

Speed = _reflection.GeneratedProtocolMessageType('Speed', (_message.Message,), dict(
  DESCRIPTOR = _SPEED,
  __module__ = 'api.stats.stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Speed)
  ))
_sym_db.RegisterMessage(Speed)


# @@protoc_insertion_point(module_scope)



from typing import Callable

from ....json_parser.game import Game
from ....generated.api.metadata import mutators_pb2

class ApiMutators:
    @staticmethod
    def create_from_game(proto_mutators: mutators_pb2.Mutators,
                        game: Game, id_creator: Callable) -> mutators_pb2.Mutators:
        proto_mutators.ball_type = game.ball_type
        proto_mutators.game_mutator_index = game.game_info.mutator_index

        return id_creator, proto_mutators
        


from logging import getLogger
from typing import Dict

import pandas as pd

from ....analysis.stats.stats import BaseStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.player_stats_pb2 import PlayerStats
from ....json_parser.game import Game


logger = getLogger(__name__)


class ControlsStat(BaseStat):

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        for player_key, stats in player_map.items():
            try:
                player_name = player_map[player_key].name

                steering_percentage = self.get_analogue_percentage("steer", data_frame, player_name)
                throttle_percentage = self.get_analogue_percentage("throttle", data_frame, player_name)
                is_keyboard = bool((steering_percentage == 0) and (throttle_percentage == 0))

                controller_stats = player_stat_map[player_key].controller
                controller_stats.is_keyboard = is_keyboard
                controller_stats.analogue_steering_input_percent = throttle_percentage
                controller_stats.analogue_throttle_input_percent = steering_percentage
            except KeyError as e:
                logger.warning('Player never pressed control %s', e)

    def get_analogue_percentage(self, column: str, data_frame: pd.DataFrame, player_name: str):
        total_frames = len(data_frame[player_name][column])
        count = (data_frame[player_name][column] == 0).sum() + (data_frame[player_name][column] == 128).sum()  + (data_frame[player_name][column] == 255).sum() + data_frame[player_name][column].isna().sum()
        return 100 - ((count * 100) / total_frames)



from typing import Callable

from ....json_parser.game import Game


class ApiGoal:
    @staticmethod
    def create_goals_from_game(game: Game, proto_goal_list, id_creator: Callable):
        for goal in game.goals:
            proto_goal = proto_goal_list.add()
            proto_goal.frame_number = goal.frame_number
            id_creator(proto_goal.player_id, goal.player_name)



import unittest

from carball.analysis.analysis_manager import AnalysisManager

from carball.json_parser.game import Game
from carball.tests.utils import run_tests_on_list, run_analysis_test_on_replay, get_raw_replays

from carball import decompile_replays


class DBTest(unittest.TestCase):
    def test_replay_attrs(self):
        local = self

        def test(replay, file_path):
            json_object = decompile_replays.decompile_replay(replay, output_path=file_path)
            game = Game()
            game.initialize(loaded_json=json_object)
            info = game.game_info
            local.assertIsNotNone(game.game_info.server_name)
            local.assertIsNotNone(game.map)
            local.assertIsNotNone(game.game_info.match_guid)

        run_tests_on_list(test)

    def test_full_replays(self):
        local = self

        def test(analysis: AnalysisManager):
            local.assertIsNotNone(analysis.get_protobuf_data())
            local.assertEqual(False, analysis.get_protobuf_data().game_metadata.is_invalid_analysis)
            for player in analysis.get_protobuf_data().players:
                ratio = (player.stats.positional_tendencies.time_in_front_ball +
                         player.stats.positional_tendencies.time_behind_ball) / player.time_in_game
                local.assertEqual(True, ratio > 0.99)
                # local.assertGreater(player.stats.positional_tendencies.time_in_front_ball, 0)
                # local.assertGreater(player.stats.positional_tendencies.time_behind_ball, 0)
                local.assertGreater(player.time_in_game, 0)
                local.assertGreater(player.stats.speed.time_at_slow_speed, 0)
                local.assertGreater(player.stats.boost.average_boost_level, 0)
                local.assertGreater(player.stats.boost.wasted_collection, -1)

        run_analysis_test_on_replay(test)

    def test_unicode_error(self):
        local = self

        def test(analysis: AnalysisManager):
            local.assertIsNotNone(analysis.get_protobuf_data())
            local.assertEqual(False, analysis.get_protobuf_data().game_metadata.is_invalid_analysis)
            for player in analysis.get_protobuf_data().players:
                ratio = (player.stats.positional_tendencies.time_in_front_ball +
                         player.stats.positional_tendencies.time_behind_ball) / player.time_in_game
                local.assertEqual(True, ratio > 0.99)
                local.assertGreater(player.stats.positional_tendencies.time_in_front_ball, 0)
                local.assertGreater(player.stats.positional_tendencies.time_behind_ball, 0)
                local.assertGreater(player.time_in_game, 0)
                local.assertGreater(player.stats.speed.time_at_slow_speed, 0)
                local.assertGreater(player.stats.boost.average_boost_level, 0)

        run_analysis_test_on_replay(test, get_raw_replays()['UNICODE_ERROR'])


if __name__ == '__main__':
    unittest.main()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/data_frame.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api import game_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/data_frame.proto',
  package='api.stats',
  serialized_pb=_b('\n\x1a\x61pi/stats/data_frame.proto\x12\tapi.stats\x1a\x0e\x61pi/game.proto:\x1d\n\ndata_frame\x12\t.api.Game\x18\x64 \x01(\x0c')
  ,
  dependencies=[game_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DATA_FRAME_FIELD_NUMBER = 100
data_frame = _descriptor.FieldDescriptor(
  name='data_frame', full_name='api.stats.data_frame', index=0,
  number=100, type=12, cpp_type=9, label=1,
  has_default_value=False, default_value=_b(""),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  options=None)

DESCRIPTOR.extensions_by_name['data_frame'] = data_frame

game_pb2.Game.RegisterExtension(data_frame)

# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/per_possession_stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/per_possession_stats.proto',
  package='api.stats',
  serialized_pb=_b('\n$api/stats/per_possession_stats.proto\x12\tapi.stats\"\xb9\x01\n\rAverageCounts\x12\r\n\x05pass_\x18\x01 \x01(\x02\x12\x0e\n\x06passed\x18\x02 \x01(\x02\x12\x0f\n\x07\x64ribble\x18\x03 \x01(\x02\x12\x1c\n\x14\x64ribble_continuation\x18\x04 \x01(\x02\x12\x0c\n\x04shot\x18\x05 \x01(\x02\x12\x0c\n\x04goal\x18\x06 \x01(\x02\x12\x0e\n\x06\x61ssist\x18\x07 \x01(\x02\x12\x10\n\x08\x61ssisted\x18\x08 \x01(\x02\x12\x0c\n\x04save\x18\t \x01(\x02\x12\x0e\n\x06\x61\x65rial\x18\n \x01(\x02\"\x85\x01\n\x12PerPossessionStats\x12\x30\n\x0e\x61verage_counts\x18\x01 \x01(\x0b\x32\x18.api.stats.AverageCounts\x12\x18\n\x10\x61verage_duration\x18\x02 \x01(\x02\x12\x14\n\x0c\x61verage_hits\x18\x03 \x01(\x02\x12\r\n\x05\x63ount\x18\x04 \x01(\x05')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_AVERAGECOUNTS = _descriptor.Descriptor(
  name='AverageCounts',
  full_name='api.stats.AverageCounts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pass_', full_name='api.stats.AverageCounts.pass_', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='passed', full_name='api.stats.AverageCounts.passed', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dribble', full_name='api.stats.AverageCounts.dribble', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dribble_continuation', full_name='api.stats.AverageCounts.dribble_continuation', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='shot', full_name='api.stats.AverageCounts.shot', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goal', full_name='api.stats.AverageCounts.goal', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='assist', full_name='api.stats.AverageCounts.assist', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='assisted', full_name='api.stats.AverageCounts.assisted', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='save', full_name='api.stats.AverageCounts.save', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aerial', full_name='api.stats.AverageCounts.aerial', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=237,
)


_PERPOSSESSIONSTATS = _descriptor.Descriptor(
  name='PerPossessionStats',
  full_name='api.stats.PerPossessionStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='average_counts', full_name='api.stats.PerPossessionStats.average_counts', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_duration', full_name='api.stats.PerPossessionStats.average_duration', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_hits', full_name='api.stats.PerPossessionStats.average_hits', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='count', full_name='api.stats.PerPossessionStats.count', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=240,
  serialized_end=373,
)

_PERPOSSESSIONSTATS.fields_by_name['average_counts'].message_type = _AVERAGECOUNTS
DESCRIPTOR.message_types_by_name['AverageCounts'] = _AVERAGECOUNTS
DESCRIPTOR.message_types_by_name['PerPossessionStats'] = _PERPOSSESSIONSTATS

AverageCounts = _reflection.GeneratedProtocolMessageType('AverageCounts', (_message.Message,), dict(
  DESCRIPTOR = _AVERAGECOUNTS,
  __module__ = 'api.stats.per_possession_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.AverageCounts)
  ))
_sym_db.RegisterMessage(AverageCounts)

PerPossessionStats = _reflection.GeneratedProtocolMessageType('PerPossessionStats', (_message.Message,), dict(
  DESCRIPTOR = _PERPOSSESSIONSTATS,
  __module__ = 'api.stats.per_possession_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.PerPossessionStats)
  ))
_sym_db.RegisterMessage(PerPossessionStats)


# @@protoc_insertion_point(module_scope)



from typing import List, ClassVar, Optional

from ..base_checks.player_check import PlayerCheck


class PlayerAttributesCheck(PlayerCheck):
    class_message: ClassVar[str] = "Player attributes must be set"

    def critical_checks(self) -> List[Optional[AssertionError]]:
        player = self.obj
        return [
            self.check(player.is_orange is not None, 'is_orange is not None'),
            self.check(player.online_id is not None, 'online_id is not None'),
            self.check(player.team is not None, 'team is not None'),
            self.check(player.name is not None, 'name is not None'),
        ]

    def major_checks(self) -> List[Optional[AssertionError]]:
        player = self.obj
        return [
                   self.check(any(player.camera_settings.values()), 'camera_settings has at least 1 value')
               ] + [
                   self.check(all(isinstance(value, int) for value in loadout.values()), 'all loadouts are ints')
                   for loadout in player.loadout
               ]



import numpy as np


FULL_BOOST_POSITIONS = np.array([
    (3000, -4100),
    (-3000, -4100),
    (3000, 0),
    (-3000, 0),
    (3000, 4100),
    (-3000, 4100),
])


def get_if_full_boost_position(position: np.array) -> bool:
    # returns 1 for full boost, 0 for small boost
    if len(position) > 2:
        position = position[0:2]

    distances_from_boosts = np.sqrt(np.square(FULL_BOOST_POSITIONS - position.values).sum(axis=1, dtype=np.float32))
    allowed_distance_from_boost = 800
    if np.any(distances_from_boosts < allowed_distance_from_boost):
        return True
    else:
        return False



import logging
from typing import List, Type

from .base_checks.base_check import BaseCheck
from .base_checks.game_check import GameCheck
from .base_checks.player_check import PlayerCheck
from .checks.ball_data_frame_check import BallDataFrameCheck
from .checks.player_attributes_check import PlayerAttributesCheck
from .checks.player_data_frame_check import PlayerDataFrameCheck
from .errors.errors import CheckErrorLevel, CheckError
from ...json_parser.game import Game

logger = logging.getLogger(__name__)

CHECKS_TO_RUN: List[Type[BaseCheck]] = [
    BallDataFrameCheck,
    PlayerAttributesCheck,
    PlayerDataFrameCheck,
]


class SanityChecker:
    """
    Checks the replays to make sure the data we are outputting are correct data.
    """

    def __init__(self, failing_level: CheckErrorLevel = CheckErrorLevel.CRITICAL):
        self.failing_level = failing_level

    def check_game(self, game: Game):
        errors: List[CheckError] = []

        for check in CHECKS_TO_RUN:
            logger.info(f'Running check: {check.__name__}')
            if issubclass(check, PlayerCheck):
                for player in game.players:
                    errors += check(player).run_check()
            elif issubclass(check, GameCheck):
                errors += check(game).run_check()

        for error in errors:
            if error.level.value >= self.failing_level.value:
                raise error

        logger.info('All checks passed.')



from typing import List

from carball.analysis.stats.possession.per_possession import PerPossessionStat
from carball.analysis.stats.ball_forward.distance_hit_ball_forward import DistanceStats
from carball.analysis.stats.boost.boost import BoostStat
from carball.analysis.stats.controls.controls import ControlsStat
from carball.analysis.stats.possession.ball_distances import BallDistanceStat
from carball.analysis.stats.possession.possession import PossessionStat
from carball.analysis.stats.possession.turnovers import TurnoverStat
from carball.analysis.stats.stats import BaseStat, HitStat
from carball.analysis.stats.tendencies.averages import Averages
from carball.analysis.stats.tendencies.hit_counts import HitCountStat
from carball.analysis.stats.tendencies.positional_tendencies import PositionalTendencies
from carball.analysis.stats.tendencies.relative_position_tendencies import RelativeTendencies
from carball.analysis.stats.tendencies.speed_tendencies import SpeedTendencies
from carball.analysis.stats.tendencies.team_tendencies import TeamTendencies


class StatsList:
    """
    Where you add any extra stats you want calculated.
    """

    @staticmethod
    def get_player_stats() -> List[BaseStat]:
        """These are stats that end up being assigned to a specific player"""
        return [BoostStat(),
                PositionalTendencies(),
                Averages(),
                BallDistanceStat(),
                ControlsStat(),
                SpeedTendencies(),
                PerPossessionStat(),
                ]

    @staticmethod
    def get_team_stats() -> List[BaseStat]:
        """These are stats that end up being assigned to a specific team"""
        return [PossessionStat(),
                TeamTendencies(),
                RelativeTendencies(),
                PerPossessionStat(),
                ]

    @staticmethod
    def get_general_stats() ->List[BaseStat]:
        """These are stats that end up being assigned to the game as a whole"""
        return [PositionalTendencies(),
                SpeedTendencies()
                ]

    @staticmethod
    def get_hit_stats() ->List[HitStat]:
        """These are stats that depend on current hit and next hit"""
        return [DistanceStats(),
                PossessionStat(),
                HitCountStat(),
                TurnoverStat()
                ]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/team.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..api import player_id_pb2
from ..api.stats import team_stats_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/team.proto',
  package='api',
  serialized_pb=_b('\n\x0e\x61pi/team.proto\x12\x03\x61pi\x1a\x13\x61pi/player_id.proto\x1a\x1a\x61pi/stats/team_stats.proto\"~\n\x04Team\x12!\n\nplayer_ids\x18\x01 \x03(\x0b\x32\r.api.PlayerId\x12\r\n\x05score\x18\x02 \x01(\x05\x12\x11\n\tis_orange\x18\x03 \x01(\x08\x12\x0c\n\x04name\x18\x04 \x01(\t\x12#\n\x05stats\x18\x05 \x01(\x0b\x32\x14.api.stats.TeamStats')
  ,
  dependencies=[player_id_pb2.DESCRIPTOR,team_stats_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TEAM = _descriptor.Descriptor(
  name='Team',
  full_name='api.Team',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='player_ids', full_name='api.Team.player_ids', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='score', full_name='api.Team.score', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_orange', full_name='api.Team.is_orange', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='api.Team.name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stats', full_name='api.Team.stats', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=72,
  serialized_end=198,
)

_TEAM.fields_by_name['player_ids'].message_type = player_id_pb2._PLAYERID
_TEAM.fields_by_name['stats'].message_type = team_stats_pb2._TEAMSTATS
DESCRIPTOR.message_types_by_name['Team'] = _TEAM

Team = _reflection.GeneratedProtocolMessageType('Team', (_message.Message,), dict(
  DESCRIPTOR = _TEAM,
  __module__ = 'api.team_pb2'
  # @@protoc_insertion_point(class_scope:api.Team)
  ))
_sym_db.RegisterMessage(Team)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/metadata/game_metadata.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api import player_id_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/metadata/game_metadata.proto',
  package='api.metadata',
  serialized_pb=_b('\n api/metadata/game_metadata.proto\x12\x0c\x61pi.metadata\x1a\x13\x61pi/player_id.proto\"7\n\tGameScore\x12\x14\n\x0cteam_0_score\x18\x01 \x01(\x05\x12\x14\n\x0cteam_1_score\x18\x02 \x01(\x05\">\n\x04Goal\x12\x14\n\x0c\x66rame_number\x18\x01 \x01(\x05\x12 \n\tplayer_id\x18\x02 \x01(\x0b\x32\r.api.PlayerId\"b\n\x04\x44\x65mo\x12\x14\n\x0c\x66rame_number\x18\x01 \x01(\x05\x12\"\n\x0b\x61ttacker_id\x18\x02 \x01(\x0b\x32\r.api.PlayerId\x12 \n\tvictim_id\x18\x03 \x01(\x0b\x32\r.api.PlayerId\"\xa4\x03\n\x0cGameMetadata\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0b\n\x03map\x18\x03 \x01(\t\x12\x0f\n\x07version\x18\x04 \x01(\x05\x12\x0c\n\x04time\x18\x05 \x01(\x03\x12\x0e\n\x06\x66rames\x18\x06 \x01(\x05\x12&\n\x05score\x18\x07 \x01(\x0b\x32\x17.api.metadata.GameScore\x12!\n\x05goals\x18\t \x03(\x0b\x32\x12.api.metadata.Goal\x12!\n\x05\x64\x65mos\x18\n \x03(\x0b\x32\x12.api.metadata.Demo\x12%\n\x0eprimary_player\x18\x0b \x01(\x0b\x32\r.api.PlayerId\x12\x0e\n\x06length\x18\x0c \x01(\x02\x12\x16\n\x0egame_server_id\x18\r \x01(\t\x12\x13\n\x0bserver_name\x18\x0e \x01(\t\x12\x12\n\nmatch_guid\x18\x0f \x01(\t\x12\x11\n\tteam_size\x18\x10 \x01(\x05\x12(\n\x08playlist\x18\x11 \x01(\x0e\x32\x16.api.metadata.Playlist\x12\x1b\n\x13is_invalid_analysis\x18\x12 \x01(\x08*\xc4\x03\n\x08Playlist\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x12\n\x0eUNRANKED_DUELS\x10\x01\x12\x14\n\x10UNRANKED_DOUBLES\x10\x02\x12\x15\n\x11UNRANKED_STANDARD\x10\x03\x12\x12\n\x0eUNRANKED_CHAOS\x10\x04\x12\x10\n\x0c\x43USTOM_LOBBY\x10\x06\x12\x17\n\x13OFFLINE_SPLITSCREEN\x10\x08\x12\x10\n\x0cRANKED_DUELS\x10\n\x12\x12\n\x0eRANKED_DOUBLES\x10\x0b\x12\x18\n\x14RANKED_SOLO_STANDARD\x10\x0c\x12\x13\n\x0fRANKED_STANDARD\x10\r\x12\x15\n\x11UNRANKED_SNOW_DAY\x10\x0f\x12\x0f\n\x0bROCKET_LABS\x10\x10\x12\x12\n\x0eUNRANKED_HOOPS\x10\x11\x12\x13\n\x0fUNRANKED_RUMBLE\x10\x12\x12\x0e\n\nTOURNAMENT\x10\x16\x12\x15\n\x11UNRANKED_DROPSHOT\x10\x17\x12\x0f\n\x0b\x41NNIVERSARY\x10\x19\x12\x10\n\x0cRANKED_HOOPS\x10\x1b\x12\x11\n\rRANKED_RUMBLE\x10\x1c\x12\x13\n\x0fRANKED_DROPSHOT\x10\x1d\x12\x13\n\x0fRANKED_SNOW_DAY\x10\x1e')
  ,
  dependencies=[player_id_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_PLAYLIST = _descriptor.EnumDescriptor(
  name='Playlist',
  full_name='api.metadata.Playlist',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRANKED_DUELS', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRANKED_DOUBLES', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRANKED_STANDARD', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRANKED_CHAOS', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUSTOM_LOBBY', index=5, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OFFLINE_SPLITSCREEN', index=6, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANKED_DUELS', index=7, number=10,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANKED_DOUBLES', index=8, number=11,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANKED_SOLO_STANDARD', index=9, number=12,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANKED_STANDARD', index=10, number=13,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRANKED_SNOW_DAY', index=11, number=15,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ROCKET_LABS', index=12, number=16,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRANKED_HOOPS', index=13, number=17,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRANKED_RUMBLE', index=14, number=18,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOURNAMENT', index=15, number=22,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRANKED_DROPSHOT', index=16, number=23,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANNIVERSARY', index=17, number=25,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANKED_HOOPS', index=18, number=27,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANKED_RUMBLE', index=19, number=28,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANKED_DROPSHOT', index=20, number=29,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RANKED_SNOW_DAY', index=21, number=30,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=716,
  serialized_end=1168,
)
_sym_db.RegisterEnumDescriptor(_PLAYLIST)

Playlist = enum_type_wrapper.EnumTypeWrapper(_PLAYLIST)
UNKNOWN = 0
UNRANKED_DUELS = 1
UNRANKED_DOUBLES = 2
UNRANKED_STANDARD = 3
UNRANKED_CHAOS = 4
CUSTOM_LOBBY = 6
OFFLINE_SPLITSCREEN = 8
RANKED_DUELS = 10
RANKED_DOUBLES = 11
RANKED_SOLO_STANDARD = 12
RANKED_STANDARD = 13
UNRANKED_SNOW_DAY = 15
ROCKET_LABS = 16
UNRANKED_HOOPS = 17
UNRANKED_RUMBLE = 18
TOURNAMENT = 22
UNRANKED_DROPSHOT = 23
ANNIVERSARY = 25
RANKED_HOOPS = 27
RANKED_RUMBLE = 28
RANKED_DROPSHOT = 29
RANKED_SNOW_DAY = 30



_GAMESCORE = _descriptor.Descriptor(
  name='GameScore',
  full_name='api.metadata.GameScore',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='team_0_score', full_name='api.metadata.GameScore.team_0_score', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='team_1_score', full_name='api.metadata.GameScore.team_1_score', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=71,
  serialized_end=126,
)


_GOAL = _descriptor.Descriptor(
  name='Goal',
  full_name='api.metadata.Goal',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='frame_number', full_name='api.metadata.Goal.frame_number', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='player_id', full_name='api.metadata.Goal.player_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=128,
  serialized_end=190,
)


_DEMO = _descriptor.Descriptor(
  name='Demo',
  full_name='api.metadata.Demo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='frame_number', full_name='api.metadata.Demo.frame_number', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='attacker_id', full_name='api.metadata.Demo.attacker_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='victim_id', full_name='api.metadata.Demo.victim_id', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=192,
  serialized_end=290,
)


_GAMEMETADATA = _descriptor.Descriptor(
  name='GameMetadata',
  full_name='api.metadata.GameMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='api.metadata.GameMetadata.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='api.metadata.GameMetadata.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='map', full_name='api.metadata.GameMetadata.map', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version', full_name='api.metadata.GameMetadata.version', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time', full_name='api.metadata.GameMetadata.time', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='frames', full_name='api.metadata.GameMetadata.frames', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='score', full_name='api.metadata.GameMetadata.score', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goals', full_name='api.metadata.GameMetadata.goals', index=7,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='demos', full_name='api.metadata.GameMetadata.demos', index=8,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='primary_player', full_name='api.metadata.GameMetadata.primary_player', index=9,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='length', full_name='api.metadata.GameMetadata.length', index=10,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='game_server_id', full_name='api.metadata.GameMetadata.game_server_id', index=11,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='server_name', full_name='api.metadata.GameMetadata.server_name', index=12,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='match_guid', full_name='api.metadata.GameMetadata.match_guid', index=13,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='team_size', full_name='api.metadata.GameMetadata.team_size', index=14,
      number=16, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='playlist', full_name='api.metadata.GameMetadata.playlist', index=15,
      number=17, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_invalid_analysis', full_name='api.metadata.GameMetadata.is_invalid_analysis', index=16,
      number=18, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=293,
  serialized_end=713,
)

_GOAL.fields_by_name['player_id'].message_type = player_id_pb2._PLAYERID
_DEMO.fields_by_name['attacker_id'].message_type = player_id_pb2._PLAYERID
_DEMO.fields_by_name['victim_id'].message_type = player_id_pb2._PLAYERID
_GAMEMETADATA.fields_by_name['score'].message_type = _GAMESCORE
_GAMEMETADATA.fields_by_name['goals'].message_type = _GOAL
_GAMEMETADATA.fields_by_name['demos'].message_type = _DEMO
_GAMEMETADATA.fields_by_name['primary_player'].message_type = player_id_pb2._PLAYERID
_GAMEMETADATA.fields_by_name['playlist'].enum_type = _PLAYLIST
DESCRIPTOR.message_types_by_name['GameScore'] = _GAMESCORE
DESCRIPTOR.message_types_by_name['Goal'] = _GOAL
DESCRIPTOR.message_types_by_name['Demo'] = _DEMO
DESCRIPTOR.message_types_by_name['GameMetadata'] = _GAMEMETADATA
DESCRIPTOR.enum_types_by_name['Playlist'] = _PLAYLIST

GameScore = _reflection.GeneratedProtocolMessageType('GameScore', (_message.Message,), dict(
  DESCRIPTOR = _GAMESCORE,
  __module__ = 'api.metadata.game_metadata_pb2'
  # @@protoc_insertion_point(class_scope:api.metadata.GameScore)
  ))
_sym_db.RegisterMessage(GameScore)

Goal = _reflection.GeneratedProtocolMessageType('Goal', (_message.Message,), dict(
  DESCRIPTOR = _GOAL,
  __module__ = 'api.metadata.game_metadata_pb2'
  # @@protoc_insertion_point(class_scope:api.metadata.Goal)
  ))
_sym_db.RegisterMessage(Goal)

Demo = _reflection.GeneratedProtocolMessageType('Demo', (_message.Message,), dict(
  DESCRIPTOR = _DEMO,
  __module__ = 'api.metadata.game_metadata_pb2'
  # @@protoc_insertion_point(class_scope:api.metadata.Demo)
  ))
_sym_db.RegisterMessage(Demo)

GameMetadata = _reflection.GeneratedProtocolMessageType('GameMetadata', (_message.Message,), dict(
  DESCRIPTOR = _GAMEMETADATA,
  __module__ = 'api.metadata.game_metadata_pb2'
  # @@protoc_insertion_point(class_scope:api.metadata.GameMetadata)
  ))
_sym_db.RegisterMessage(GameMetadata)


# @@protoc_insertion_point(module_scope)



import unittest

from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_specific_replays, get_raw_replays


class KickOffTest(unittest.TestCase):

    def test_0_kickoffs(self):

        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            self.assertEqual(len(kickoffs), 0)

        run_analysis_test_on_replay(test, get_raw_replays()["KICKOFF_NO_TOUCH"])

    def test_3_kickoffs(self):
        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            kickoffs = proto_game.game_stats.kickoffs
            self.assertEqual(len(kickoffs), 3)

        run_analysis_test_on_replay(test, get_raw_replays()["3_KICKOFFS"])


if __name__ == '__main__':
    unittest.main()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/metadata/mutators.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/metadata/mutators.proto',
  package='api.metadata',
  serialized_pb=_b('\n\x1b\x61pi/metadata/mutators.proto\x12\x0c\x61pi.metadata\"Q\n\x08Mutators\x12)\n\tball_type\x18\x01 \x01(\x0e\x32\x16.api.metadata.BallType\x12\x1a\n\x12game_mutator_index\x18\x02 \x01(\x05*M\n\x08\x42\x61llType\x12\x0b\n\x07\x44\x45\x46\x41ULT\x10\x01\x12\x0e\n\nBASKETBALL\x10\x02\x12\x08\n\x04PUCK\x10\x03\x12\x0c\n\x08\x43UBEBALL\x10\x04\x12\x0c\n\x08\x42REAKOUT\x10\x05')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_BALLTYPE = _descriptor.EnumDescriptor(
  name='BallType',
  full_name='api.metadata.BallType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DEFAULT', index=0, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BASKETBALL', index=1, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PUCK', index=2, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUBEBALL', index=3, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BREAKOUT', index=4, number=5,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=128,
  serialized_end=205,
)
_sym_db.RegisterEnumDescriptor(_BALLTYPE)

BallType = enum_type_wrapper.EnumTypeWrapper(_BALLTYPE)
DEFAULT = 1
BASKETBALL = 2
PUCK = 3
CUBEBALL = 4
BREAKOUT = 5



_MUTATORS = _descriptor.Descriptor(
  name='Mutators',
  full_name='api.metadata.Mutators',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ball_type', full_name='api.metadata.Mutators.ball_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='game_mutator_index', full_name='api.metadata.Mutators.game_mutator_index', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=45,
  serialized_end=126,
)

_MUTATORS.fields_by_name['ball_type'].enum_type = _BALLTYPE
DESCRIPTOR.message_types_by_name['Mutators'] = _MUTATORS
DESCRIPTOR.enum_types_by_name['BallType'] = _BALLTYPE

Mutators = _reflection.GeneratedProtocolMessageType('Mutators', (_message.Message,), dict(
  DESCRIPTOR = _MUTATORS,
  __module__ = 'api.metadata.mutators_pb2'
  # @@protoc_insertion_point(class_scope:api.metadata.Mutators)
  ))
_sym_db.RegisterMessage(Mutators)


# @@protoc_insertion_point(module_scope)



import os
import logging
import platform as pt

from carball.analysis.analysis_manager import AnalysisManager
from carball.extras.per_goal_analysis import PerGoalAnalysis
from carball.json_parser.sanity_check.sanity_check import SanityChecker
from carball.json_parser.game import Game
from carball.controls.controls import ControlsCreator
from carball.rattletrap import run_rattletrap


logger = logging.getLogger(__name__)
BASE_DIR = os.path.dirname(__file__)


def decompile_replay(replay_path, output_path: str=None, overwrite=True):
    """
    Takes a path to the replay and outputs the json of that replay.

    :param replay_path: Path to a specific replay.
    :param output_path: The output path of rattletrap.
    :param overwrite: True if we should recreate the json even if it already exists.
    :return: The json created from rattle trap.
    """
    return run_rattletrap.decompile_replay(replay_path, output_path)


def analyze_replay_file(replay_path: str, output_path: str=None, overwrite=True, controls: ControlsCreator=None,
                        sanity_check: SanityChecker=None, analysis_per_goal=False):
    """
    Decompile and analyze a replay file.

    :param replay_path: Path to replay file
    :param output_path: Path to write JSON
    :param overwrite: If to overwrite JSON (suggest True if speed is not an issue)
    :param controls: Generate controls from the replay using our best guesses (ALPHA)
    :param sanity_check: Run sanity check to make sure we analyzed correctly (BETA)
    :param analysis_per_goal: Runs the analysis per a goal instead of the replay as a whole
    :return: AnalysisManager of game with analysis.
    """
    _json = decompile_replay(replay_path, output_path=output_path, overwrite=overwrite)
    game = Game()
    game.initialize(loaded_json=_json)
    # get_controls(game)  # TODO: enable and optimise.
    if sanity_check is not None:
        sanity_check.check_game(game)
    if analysis_per_goal:
        analysis = PerGoalAnalysis(game)
    else:
        analysis = AnalysisManager(game)
    analysis.create_analysis()

    if controls is not None:
        controls.get_controls(game)

    return analysis


if __name__ == '__main__':
    from carball.tests.analysis_test import __test_replays
    __test_replays(BASE_DIR)



import gzip
import io
import struct
import numpy as np


def write_array_to_gzip(file_path, array: np.ndarray):
    with gzip.open(file_path, 'wb') as f:
        write_array_to_file(f, array)


def write_array_to_file(game_file, array: np.ndarray):
    """
    :param game_file: This is the file that the array will be written to.
    :param array: A numpy array of any size.
    """
    bytes = convert_numpy_array(array)
    size_of_bytes = len(bytes.getvalue())
    game_file.write(struct.pack('i', size_of_bytes))
    game_file.write(bytes.getvalue())


def convert_numpy_array(numpy_array: np.ndarray):
    """
    Converts a numpy array into compressed bytes
    :param numpy_array: An array that is going to be converted into bytes
    :return: A BytesIO object that contains compressed bytes
    """
    compressed_array = io.BytesIO()  # np.savez_compressed() requires a file-like object to write to
    np.save(compressed_array, numpy_array, allow_pickle=True, fix_imports=False)
    return compressed_array


def read_array_from_file(file):
    chunk = file.read(4)
    numpy_array, num_bytes = get_array(file, chunk)
    return numpy_array


def get_array(file, chunk):
    """
    Gets a compressed numpy array from a file.
    Throws an EOFError if it has problems loading the data.
    :param file: The file that is being read
    :param chunk: A chunk representing a single number, this will be the number of bytes the array takes up.
    :return: A numpy array
    """
    try:
        starting_byte = struct.unpack('i', chunk)[0]
    except struct.error:
        raise EOFError('Struct error')
    numpy_bytes = file.read(starting_byte)
    fake_file = io.BytesIO(numpy_bytes)
    try:
        result = np.load(fake_file, fix_imports=False)
    except OSError:
        raise EOFError('NumPy parsing error')
    return result, starting_byte



import pandas as pd
import numpy as np

position_column_names = ['pos_x', 'pos_y', 'pos_z']
positional_columns = ['pos_x', 'pos_y', 'pos_z', 'rot_x', 'rot_y', 'rot_z']

def get_player_ball_displacements(data_frame: pd.DataFrame, player_name: str) -> pd.DataFrame:
    player_df = data_frame[player_name]
    ball_df = data_frame['ball']

    return get_position_displacements(player_df, ball_df)


def get_position_displacements(data_frame1: pd.DataFrame, data_frame2: pd.DataFrame) -> pd.DataFrame:
    result = data_frame1[position_column_names] - data_frame2[position_column_names]
    return result


def get_distance_from_displacements(data_frame: pd.DataFrame) -> pd.Series:
    positions = data_frame[position_column_names]

    summed = (positions ** 2).sum(axis=1, skipna=False)
    return np.sqrt(summed)



from typing import Dict

import pandas as pd

from ....analysis.stats.stats import HitStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.events_pb2 import Hit
from ....json_parser.game import Game


class DistanceStats(HitStat):

    def initialize_hit_stat(self, game: Game, player_map: Dict[str, Player], data_frame: pd.DataFrame):
        pass

    def calculate_next_hit_stat(self, game: Game, proto_game: game_pb2.Game, saltie_hit: Hit, next_saltie_hit: Hit,
                                player_map: Dict[str, Player], hit_index: int):
        player = player_map[saltie_hit.player_id.id]
        hit_distance_y = next_saltie_hit.ball_data.pos_y - saltie_hit.ball_data.pos_y

        if player.is_orange:
            hit_distance_y *= -1

        if hit_distance_y > 0:
            player.stats.distance.ball_hit_forward += hit_distance_y
        if hit_distance_y < 0:
            player.stats.distance.ball_hit_backward += abs(hit_distance_y)



from typing import Optional

from ....json_parser.player import Player
from .base_check import BaseCheck


class PlayerCheck(BaseCheck[Player]):

    def check(self, condition: bool, message: str) -> Optional[AssertionError]:
        return super().check(condition, f'{message} ({self.obj.name})')



import os
import tempfile
from typing import Callable

import requests
from carball.decompile_replays import analyze_replay_file


def download_replay_discord(url):
    file = requests.get(url, stream=True)
    replay = file.raw.data
    return replay


def save_locally(replay_object):
    fd, file_path = tempfile.mkstemp()
    with open(file_path, 'wb') as f:
        f.write(replay_object)

    os.close(fd)
    return file_path


def run_tests_on_list(unit_test_func: Callable, replay_list=None, answers=None):
    if replay_list is None:
        replay_list = get_complex_replay_list()

    for index in range(len(replay_list)):
        replay_url = replay_list[index]
        print('running test on replay: ' + replay_url[replay_url.rfind('/') + 1:])
        answer = answers[index] if answers is not None and index < len(answers) else None
        run_replay(replay_url, unit_test_func, answer=answer)


def run_replay(url, unit_test_func: Callable, answer=None):
    """
    Runs the replay with the file downloaded locally then deletes the file.
    :param url:
    :param unit_test_func:
    :param answer: data that can be passed to the replay to help judge it
    :return:
    """

    replay = download_replay_discord(url)
    file = save_locally(replay)

    fd, file_path = tempfile.mkstemp()
    os.close(fd)
    if answer is not None:
        unit_test_func(file, file_path, answer)
    else:
        unit_test_func(file, file_path)
    os.remove(file)
    os.remove(file_path)


def run_analysis_test_on_replay(unit_test_func: Callable, replay_list=None, answers=None):
    """
    :param unit_test_func: Called with an AnalysisManager
    :param replay_list: list of replay urls
    :return:
    """

    def wrapper(replay_file_path, json_file_path, answer=None):
        analysis_manager = analyze_replay_file(replay_file_path)
        if answer is not None:
            unit_test_func(analysis_manager, answer)
        else:
            unit_test_func(analysis_manager)

    run_tests_on_list(wrapper, replay_list, answers=answers)


def get_complex_replay_list():
    """
    For full replays that have crashed or failed to be converted
    :return:
    """
    return [
        'https://cdn.discordapp.com/attachments/493849514680254468/496153554977816576/BOTS_JOINING_AND_LEAVING.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/496153569981104129/BOTS_NO_POSITION.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/496153605074845734/ZEROED_STATS.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/496180938968137749/FAKE_BOTS_SkyBot.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/497149910999891969/NEGATIVE_WASTED_COLLECTION.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/497191273619259393/WASTED_BOOST_WHILE_SUPER_SONIC.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/501630263881760798/OCE_RLCS_7_CARS.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/561300088400379905/crossplatform_party.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/563036945635082260/PARTY_LEADER_SYSTEM_ID_0.replay',
    ]


def get_raw_replays():
    """
    Partial replays for specific values.
    :return:
    """
    return {
        "0_JUMPS": ["https://cdn.discordapp.com/attachments/493849514680254468/495735116895748096/0_JUMPS.replay"],
        "0_SAVES": ["https://cdn.discordapp.com/attachments/493849514680254468/495735137133264897/0_SAVES.replay"],
        "1_AERIAL": ["https://cdn.discordapp.com/attachments/493849514680254468/495735149133168651/1_AERIAL.replay"],
        "1_DEMO": ["https://cdn.discordapp.com/attachments/493849514680254468/495735163117109249/1_DEMO.replay"],
        "1_DOUBLE_JUMP": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735176761049103/1_DOUBLE_JUMP.replay"],
        "1_EPIC_SAVE": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735191537713153/1_EPIC_SAVE.replay"],
        "1_JUMP": ["https://cdn.discordapp.com/attachments/493849514680254468/495735203323576321/1_JUMP.replay"],
        "1_NORMAL_SAVE_FROM_SHOT_TOWARD_POST": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735215767945234/1_NORMAL_SAVE.replay"],

        # Boost
        "12_BOOST_PAD_0_USED": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735228422291456/12_BOOST_PAD_0_USED.replay"],
        "12_BOOST_PAD_45_USED": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735240321662986/12_BOOST_PAD_45_USED.replay"],
        "100_BOOST_PAD_0_USED": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735252233224193/100_BOOST_PAD_0_USED.replay"],
        "100_BOOST_PAD_100_USED": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735264262488065/100_BOOST_PAD_100_USED.replay"],
        "NO_BOOST_PAD_0_USED": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735276338020372/NO_BOOST_PAD_0_USED.replay"],
        "NO_BOOST_PAD_33_USED": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735288254169109/NO_BOOST_PAD_33_USED.replay"],
        "12_AND_100_BOOST_PADS_0_USED": [
            "https://cdn.discordapp.com/attachments/493849514680254468/496071113768697859/12_AND_100_BOOST_PADS_0_USED.replay"],
        "WASTED_BOOST_WHILE_SUPER_SONIC": [
            "https://cdn.discordapp.com/attachments/493849514680254468/497191273619259393/WASTED_BOOST_WHILE_SUPER_SONIC.replay"],
        "CALCULATE_USED_BOOST_WITH_DEMO": [
            "https://cdn.discordapp.com/attachments/493849514680254468/497189284651204609/CALCULATE_USED_BOOST_WITH_DEMO.replay"],
        "CALCULATE_USED_BOOST_DEMO_WITH_FLIPS": [
            "https://cdn.discordapp.com/attachments/493849514680254468/497189968397991937/CALCULATE_USED_BOOST_DEMO_WITH_FLIPS.replay"],
        "MORE_THAN_100_BOOST": [
            "https://cdn.discordapp.com/attachments/493849514680254468/497190406472204288/MORE_THAN_100_BOOST.replay"],
        "USE_BOOST_AFTER_GOAL": [
            "https://cdn.discordapp.com/attachments/493849514680254468/497190907309850634/USE_BOOST_AFTER_GOAL.replay"
        ],
        "FEATHERING_34x100_BO0ST_USED": [
            "https://cdn.discordapp.com/attachments/493849514680254468/499640872313290763/FEATHERING_34_X_100_BOOSTS_USED.replay"
        ],

        # Kickoffs
        "STRAIGHT_KICKOFF_GOAL": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495735301604376576/Straight_Kickoff_Goal.replay"],
        "KICKOFF_NO_TOUCH": [
            "https://cdn.discordapp.com/attachments/493849514680254468/496034430943756289/NO_KICKOFF.replay"],
        "3_KICKOFFS": [
            "https://cdn.discordapp.com/attachments/493849514680254468/496034443442782208/3_KICKOFFS_4_SHOTS.replay"],

        # hits
        "4_SHOTS": [
            "https://cdn.discordapp.com/attachments/493849514680254468/496034443442782208/3_KICKOFFS_4_SHOTS.replay"],
        "KICKOFF_3_HITS": [
            "https://cdn.discordapp.com/attachments/493849514680254468/496072257928691743/KICKOFF_3_HITS.replay"],
        "MID_AIR_PASS": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495887162928267314/MID_AIR_PASS_GOAL.replay"],
        "HIGH_AIR_PASS": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495887164425633802/HIGH_AIR_PASS_GOAL.replay"],
        "GROUND_PASS": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495887165570678794/GROUNDED_PASS_GOAL.replay"],
        "PINCH_GROUND": [
            "https://cdn.discordapp.com/attachments/493849514680254468/495887167932071947/PINCH_GROUNDED_GOAL.replay"],
        "DEFAULT_3_ON_3_AROUND_58_HITS": [
            "https://cdn.discordapp.com/attachments/493849514680254468/496820586811621387/DEFAULT_3_ON_3_AROUND_58_HITS.replay"],

        # parties
        "PLAY_STATION_ONLY_PARTY": [
            'https://cdn.discordapp.com/attachments/493849514680254468/563457368193761301/PLAY_STATION_ONLY_PARTY.replay'],

        "XBOX_PARTY": [
            "https://cdn.discordapp.com/attachments/493849514680254468/563831620222844928/XBOX_PARTY.replay"
        ],
        # error cases
        "UNICODE_ERROR": [
            "https://cdn.discordapp.com/attachments/493849514680254468/493880540802449462/UnicodeEncodeError.replay"],
        "CROSSPLATFORM_PARTY_LEADER_ERROR": [
            "https://cdn.discordapp.com/attachments/493849514680254468/561300088400379905/crossplatform_party.replay"],
        "PARTY_LEADER_SYSTEM_ID_0_ERROR": [
            'https://cdn.discordapp.com/attachments/493849514680254468/563036945635082260/PARTY_LEADER_SYSTEM_ID_0.replay'],
    }


def get_specific_replays():
    raw_map = get_raw_replays()
    return {
        # BOOSTS
        "0_BOOST_COLLECTED": raw_map["NO_BOOST_PAD_0_USED"] + raw_map["NO_BOOST_PAD_33_USED"] + raw_map[
            "KICKOFF_NO_TOUCH"],
        "1_SMALL_PAD": raw_map["12_BOOST_PAD_0_USED"] + raw_map["12_BOOST_PAD_45_USED"],
        "1_LARGE_PAD": raw_map["100_BOOST_PAD_0_USED"] + raw_map["100_BOOST_PAD_100_USED"],
        "0_BOOST_USED": raw_map["12_BOOST_PAD_0_USED"] + raw_map["100_BOOST_PAD_0_USED"] + raw_map[
            "NO_BOOST_PAD_0_USED"] + raw_map["KICKOFF_NO_TOUCH"],
        "BOOST_USED": raw_map["12_BOOST_PAD_45_USED"] +
                      raw_map["100_BOOST_PAD_100_USED"] +
                      raw_map["NO_BOOST_PAD_33_USED"] +
                      raw_map["CALCULATE_USED_BOOST_WITH_DEMO"] +
                      raw_map["CALCULATE_USED_BOOST_DEMO_WITH_FLIPS"] +
                      raw_map["USE_BOOST_AFTER_GOAL"] +
                      raw_map["WASTED_BOOST_WHILE_SUPER_SONIC"],
        "BOOST_FEATHERED": raw_map["MORE_THAN_100_BOOST"] + raw_map["FEATHERING_34x100_BO0ST_USED"],
        "BOOST_WASTED_USAGE": raw_map["WASTED_BOOST_WHILE_SUPER_SONIC"],
        "BOOST_WASTED_COLLECTION": raw_map["MORE_THAN_100_BOOST"],
        # HITS
        "HITS": raw_map["4_SHOTS"] + raw_map["KICKOFF_3_HITS"] + raw_map["12_BOOST_PAD_45_USED"] +
                raw_map["MID_AIR_PASS"] + raw_map["HIGH_AIR_PASS"] + raw_map["GROUND_PASS"] +
                raw_map["1_NORMAL_SAVE_FROM_SHOT_TOWARD_POST"] + raw_map["1_EPIC_SAVE"] + raw_map["1_AERIAL"] +
                raw_map["DEFAULT_3_ON_3_AROUND_58_HITS"],
        # + raw_map["PINCH_GROUND"],  TODO: Fix pinches to create 2 hits 1 for each person on same frame
        "SHOTS": raw_map["4_SHOTS"] + raw_map["12_BOOST_PAD_45_USED"] +
                 raw_map["1_EPIC_SAVE"] + raw_map["1_NORMAL_SAVE_FROM_SHOT_TOWARD_POST"],
        "PASSES": raw_map["MID_AIR_PASS"] + raw_map["HIGH_AIR_PASS"] + raw_map["GROUND_PASS"],
        "AERIALS": raw_map["1_EPIC_SAVE"] + raw_map["1_AERIAL"] + raw_map["HIGH_AIR_PASS"] + raw_map["MID_AIR_PASS"],
        "SAVES": raw_map["1_EPIC_SAVE"] + raw_map["1_NORMAL_SAVE_FROM_SHOT_TOWARD_POST"]
    }


def get_specific_answers():
    specific_replays = get_specific_replays()
    return {
        # Boost
        "0_BOOST_USED": [0] * len(specific_replays["0_BOOST_USED"]),
        "BOOST_USED": [45, 100, 33, 33.33 + 33.33 + 12.15, 33.33, 33.33, 0],
        "BOOST_WASTED_USAGE": [33.33],
        "BOOST_WASTED_COLLECTION": [6.2],
        "BOOST_FEATHERED": [100.0, 3490.0],
        # Hits
        "HITS": [4, 3, 1, 2, 9, 2, 4, 4, 4, 50],
        "SHOTS": [3, 0, 2, 1],
        "PASSES": [1, 1, 1],
        "AERIALS": [0, 1, 2, 0],
        "SAVES": [1, 0],
    }


def assertNearlyEqual(self, a, b, percent=2.0, msg=None):
    if abs(a - b) > abs(percent / 100.0 * min(abs(a), abs(b))):
        if msg is None:
            self.fail("The given numbers %s and %s are not within %s percent of each other." % (a, b, percent))
        else:
            self.fail(msg)



import numpy as np
import pandas as pd

T_r = -36.07956616966136  # torque coefficient for roll
T_p = -12.14599781908070  # torque coefficient for pitch
T_y = 8.91962804287785  # torque coefficient for yaw
D_r = -4.47166302201591  # drag coefficient for roll
D_p = -2.798194258050845  # drag coefficient for pitch
D_y = -1.886491900437232  # drag coefficient for yaw


def predict_user_inputs(ang_vels, rotations, deltas):
    delta_omega = -ang_vels.diff(-1)
    delta_omega = delta_omega[(delta_omega.T != 0).any()]  # Filter all-zero rows.
    tau = delta_omega.divide(deltas[delta_omega.index], axis='index')
    tau_vectors = np.expand_dims(tau.values, 2)

    rotation_matrices = get_rotation_matrices(rotations.loc[delta_omega.index])
    inverse_rotation_matrices: pd.Series = np.transpose(rotation_matrices)
    inverse_rotation_array = np.stack(inverse_rotation_matrices.values)

    tau_local = np.matmul(inverse_rotation_array, tau_vectors)

    ang_vel_vectors = np.expand_dims(ang_vels.loc[delta_omega.index].values, 2)
    omega_local = np.matmul(inverse_rotation_array, ang_vel_vectors)

    omega_and_tau_locals = np.concatenate([omega_local, tau_local], axis=1)

    rhs_and_omega = np.apply_along_axis(get_rhs_and_omega, 1, omega_and_tau_locals)

    u = np.apply_along_axis(get_u, 1, rhs_and_omega)
    controls_data_frame = pd.DataFrame(data=np.squeeze(u, 2),
                                       index=delta_omega.index,
                                       columns=['predicted_input_roll', 'predicted_input_pitch', 'predicted_input_yaw'])

    controls_data_frame.clip(-1, 1, inplace=True)
    return controls_data_frame


def get_rotation_matrices(rotations: pd.Series) -> pd.Series:
    pitch = rotations.rot_x
    yaw = rotations.rot_y
    roll = rotations.rot_z

    cos_roll = np.cos(roll).rename('cos_roll')
    sin_roll = np.sin(roll).rename('sin_roll')
    cos_pitch = np.cos(pitch).rename('cos_pitch')
    sin_pitch = np.sin(pitch).rename('sin_pitch')
    cos_yaw = np.cos(yaw).rename('cos_yaw')
    sin_yaw = np.sin(yaw).rename('sin_yaw')

    components: pd.DataFrame = pd.concat([cos_roll, sin_roll, cos_pitch, sin_pitch, cos_yaw, sin_yaw], axis=1)

    rotation_matrix = components.apply(get_rotation_matrix_from_row, axis=1, result_type='reduce')
    return rotation_matrix


def get_rotation_matrix_from_row(components: pd.Series) -> np.array:
    cos_roll, sin_roll, cos_pitch, sin_pitch, cos_yaw, sin_yaw = components.values
    rotation_matrix = np.array(
        [[cos_pitch * cos_yaw, cos_yaw * sin_pitch * sin_roll - cos_roll * sin_yaw,
          -cos_roll * cos_yaw * sin_pitch - sin_roll * sin_yaw],
         [cos_pitch * sin_yaw, sin_yaw * sin_pitch * sin_roll + cos_roll * cos_yaw,
          -cos_roll * sin_yaw * sin_pitch + sin_roll * cos_yaw],
         [sin_pitch, -cos_pitch * sin_roll, cos_pitch * cos_roll]])
    return rotation_matrix


def get_rhs_and_omega(omega_and_tau):
    omega = omega_and_tau[:3]
    tau = omega_and_tau[3:]
    return np.array([
        tau[0] - D_r * omega[0],
        tau[1] - D_p * omega[1],
        tau[2] - D_y * omega[2],
        omega[0],
        omega[1],
        omega[2]
    ])


def get_u(rhs_and_omega):
    rhs = rhs_and_omega[:3]
    omega = rhs_and_omega[3:]
    return np.array([
        rhs[0] / T_r,
        rhs[1] / (T_p + np.sign(rhs[1]) * omega[1] * D_p),
        rhs[2] / (T_y - np.sign(rhs[2]) * omega[2] * D_y)
    ])



from typing import Callable
from ....generated.api import player_pb2
from ....json_parser.player import Player
from .ApiPlayerCameraSettings import ApiPlayerCameraSettings
from .ApiPlayerLoadout import ApiPlayerLoadout


class ApiPlayer:

    @staticmethod
    def create_from_player(proto_player: player_pb2.Player, player: Player, id_creator: Callable):
        ApiPlayerCameraSettings.create_from_player(proto_player.camera_settings, player)
        ApiPlayerLoadout.create_from_player(proto_player.loadout, player)

        id_creator(proto_player.id, player.name)
        if player.name is not None:
            proto_player.name = player.name
        if player.title is not None:
            proto_player.title_id = player.title
        if player.score is not None:
            proto_player.score = player.score
        if player.goals is not None:
            proto_player.goals = player.goals
        if player.assists is not None:
            proto_player.assists = player.assists
        if player.saves is not None:
            proto_player.saves = player.saves
        if player.shots is not None:
            proto_player.shots = player.shots
        if player.is_orange is not None:
            proto_player.is_orange = player.is_orange
        if player.party_leader is not None:
            proto_player.party_leader.id = player.party_leader
        if player.is_bot is not None:
            proto_player.is_bot = player.is_bot

        return proto_player



from enum import Enum

import pandas as pd
import numpy as np

from ..simulator.map_constants import MAP_Y


class FieldType(Enum):
    STANDARD = 1

STANDARD_FIELD_LENGTH_HALF = 5120
STANDARD_FIELD_WIDTH_HALF = 4096
STANDARD_GOAL_WIDTH_HALF = 893

HEIGHT_0_BALL_LIM = 95  # Height of car when on ground
HEIGHT_0_LIM = 20  # Height of car when on ground
HEIGHT_1_LIM = 840  # Goal height

MAP_THIRD = MAP_Y / 6

NEUTRAL_ZONE = MAP_Y / 20


class FieldConstants:

    field_type = FieldType.STANDARD

    corner = np.array([STANDARD_FIELD_WIDTH_HALF - STANDARD_GOAL_WIDTH_HALF,
                       STANDARD_FIELD_LENGTH_HALF - STANDARD_GOAL_WIDTH_HALF])
    near_wall = np.array([STANDARD_FIELD_WIDTH_HALF - STANDARD_GOAL_WIDTH_HALF / 2,
                          STANDARD_FIELD_LENGTH_HALF - STANDARD_GOAL_WIDTH_HALF / 2])
    rectangle_lower = np.array(-near_wall)
    rectangle_higher = np.array(near_wall)

    def __init__(self, field_type=FieldType.STANDARD):
        self.field_type = field_type

    """
    All functions here assume the car is on the blue team and the game is played on a standard map.
    return: Boolean series that can be used to index the original data_frame to sum deltas with.
    """

    def get_neutral_zone(self, player_data_frame, **kwargs):
        return self.abs(player_data_frame.pos_y) < NEUTRAL_ZONE

    def get_half_0(self, player_data_frame, **kwargs):
        return player_data_frame.pos_y < 0

    def get_half_1(self, player_data_frame, **kwargs):
        return player_data_frame.pos_y > 0

    def get_third_0(self, player_data_frame, **kwargs):
        return player_data_frame.pos_y < -MAP_THIRD

    def get_third_1(self, player_data_frame, **kwargs):
        return (-MAP_THIRD < player_data_frame.pos_y) & (player_data_frame.pos_y < MAP_THIRD)

    def get_third_2(self, player_data_frame, **kwargs):
        return player_data_frame.pos_y > MAP_THIRD

    def get_height_0(self, player_data_frame, **kwargs):
        return player_data_frame.pos_z < HEIGHT_0_LIM

    def get_height_0_ball(self, player_data_frame, **kwargs):
        return player_data_frame.pos_z < HEIGHT_0_BALL_LIM

    def get_height_1(self, player_data_frame, **kwargs):
        return (HEIGHT_0_LIM < player_data_frame.pos_z) & (player_data_frame.pos_z < HEIGHT_1_LIM)

    def get_height_2(self, player_data_frame, **kwargs):
        return player_data_frame.pos_z > HEIGHT_1_LIM

    def get_ball_0(self, player_data_frame, ball_data_frame):
        """Ball is closer to goal 0 than player"""
        return player_data_frame.pos_y < ball_data_frame.pos_y

    def get_ball_1(self, player_data_frame, ball_data_frame):
        """Ball is closer to goal 1 than player"""
        return player_data_frame.pos_y > ball_data_frame.pos_y

    def get_wall_time(self, player_data_frame, **kwargs):
        return ~((self.rectangle_lower[0] <= player_data_frame.pos_x) &
                 (player_data_frame.pos_x <= self.rectangle_higher[0]) &
                 (self.rectangle_lower[1] <= player_data_frame.pos_y) &
                 (player_data_frame.pos_y <= self.rectangle_higher[1]))

    def get_corner_time(self, player_data_frame, **kwargs):
        return (((player_data_frame.pos_x >= self.corner[0]) |
                (player_data_frame.pos_x <= -self.corner[0])) &
                ((player_data_frame.pos_y >= self.corner[1]) |
                (player_data_frame.pos_y <= - self.corner[1])))

    def abs(self, value):
        if value is pd.DataFrame:
            return value.abs()
        else:
            return abs(value)



import logging

import numpy as np
import pandas as pd

from ..json_parser.bots import get_bot_map, get_online_id_for_bot
from .boost import get_if_full_boost_position

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from .team import Team

logger = logging.getLogger(__name__)
bot_map = get_bot_map()


class Player:

    def __init__(self):
        self.name = None
        self.online_id = None
        self.team = None  # using team class. set later.
        self.is_orange = None
        self.score = None
        self.goals = None
        self.assists = None
        self.saves = None
        self.shots = None
        self.is_bot = None

        self.frame_data = {}

        self.camera_settings = {}
        self.loadout = []

        self.data = None
        self.boosts = None
        self.demos = None

        self.title = None
        self.total_xp = None
        self.steering_sensitivity = None

        self.party_leader = None

    def __repr__(self):
        if self.team:
            return '%s: %s on %s' % (self.__class__.__name__, self.name, self.team)
        else:
            return '%s: %s' % (self.__class__.__name__, self.name)

    def create_from_actor_data(self, actor_data: dict, teams: List['Team']):
        self.name = actor_data['name']
        if 'Engine.PlayerReplicationInfo:bBot' in actor_data and actor_data['Engine.PlayerReplicationInfo:bBot']:
            self.is_bot = True
            self.online_id = get_online_id_for_bot(bot_map, self)

        else:
            actor_type = list(actor_data["Engine.PlayerReplicationInfo:UniqueId"]['unique_id']['remote_id'].keys())[0]
            self.online_id = actor_data["Engine.PlayerReplicationInfo:UniqueId"]['unique_id']['remote_id'][actor_type]
        try:
            self.score = actor_data["TAGame.PRI_TA:MatchScore"]
        except KeyError:
            logger.warning('Score is not found for player')
        team_actor_id = actor_data["Engine.PlayerReplicationInfo:Team"]
        if team_actor_id == -1:
            # if they leave at the end
            team_actor_id = actor_data['team']
        for team in teams:
            if team.actor_id == team_actor_id:
                self.is_orange = team.is_orange
        assert self.is_orange is not None, 'Cannot find team for player: %s' % self.name
        self.score = actor_data.get("TAGame.PRI_TA:MatchScore", None)
        self.goals = actor_data.get("TAGame.PRI_TA:MatchGoals", None)
        self.assists = actor_data.get("TAGame.PRI_TA:MatchAssists", None)
        self.saves = actor_data.get("TAGame.PRI_TA:MatchSaves", None)
        self.shots = actor_data.get("TAGame.PRI_TA:MatchShots", None)
        self.parse_actor_data(actor_data)

        logger.info('Created Player from actor_data: %s' % self)
        return self

    def parse_player_stats(self, player_stats: dict):
        self.name = player_stats["Name"]["value"]["str"]
        self.online_id = str(player_stats["OnlineID"]["value"]["q_word"])
        self.is_orange = bool(player_stats["Team"]["value"]["int"])
        self.score = player_stats["Score"]["value"]["int"]
        self.goals = player_stats["Goals"]["value"]["int"]
        self.assists = player_stats["Assists"]["value"]["int"]
        self.saves = player_stats["Saves"]["value"]["int"]
        self.shots = player_stats["Shots"]["value"]["int"]
        self.is_bot = bool(player_stats["bBot"]["value"]["bool"])

        logger.info('Created Player from stats: %s' % self)
        if self.is_bot or self.online_id == '0' or self.online_id == 0:
            self.online_id = get_online_id_for_bot(bot_map, self)

        return self

    def get_camera_settings(self, camera_data: dict):
        self.camera_settings['field_of_view'] = camera_data.get('fov', None)
        self.camera_settings['height'] = camera_data.get('height', None)
        self.camera_settings['pitch'] = camera_data.get('angle', None)
        self.camera_settings['distance'] = camera_data.get('distance', None)
        self.camera_settings['stiffness'] = camera_data.get('stiffness', None)
        self.camera_settings['swivel_speed'] = camera_data.get('swivel_speed', None)
        self.camera_settings['transition_speed'] = camera_data.get('transition_speed', None)

        for key, value in self.camera_settings.items():
            if value is None:
                logger.warning('Could not find ' + key + ' in camera settings for ' + self.name)
        logger.info('Camera settings for %s: %s' % (self.name, self.camera_settings))

    def parse_actor_data(self, actor_data: dict):
        """
        Adds stuff not found in PlayerStats metadata.
        PlayerStats is a better source of truth - as actor_data might not have been updated (e.g. for last assist)

        :param actor_data:
        :return:
        """
        self.get_loadout(actor_data)
        self.party_leader = actor_data.get('TAGame.PRI_TA:PartyLeader', None)
        try:
            if self.party_leader is not None and \
                    'party_leader' in self.party_leader and \
                    'id' in self.party_leader['party_leader']:
                leader_actor_type = list(self.party_leader['party_leader']['id'][0].keys())[0]
                self.party_leader = str(self.party_leader['party_leader']['id'][0][leader_actor_type])
            else:
                self.party_leader = None
        except KeyError:
            logger.warning('Could not set player party leader for:', self.name)
            self.party_leader = None
        self.title = actor_data.get('TAGame.PRI_TA:Title', None)
        self.total_xp = actor_data.get('TAGame.PRI_TA:TotalXP', None)
        self.steering_sensitivity = actor_data.get('TAGame.PRI_TA:SteeringSensitivity', None)
        return self

    def get_loadout(self, actor_data: dict):
        if "TAGame.PRI_TA:ClientLoadouts" in actor_data:  # new version (2 loadouts)
            loadout_data = actor_data["TAGame.PRI_TA:ClientLoadouts"]["loadouts"]
        else:
            loadout_data = {'0': actor_data["TAGame.PRI_TA:ClientLoadout"]["loadout"]}
        for loadout_name, _loadout in loadout_data.items():
            self.loadout.append({
                'version': _loadout.get('version', None),
                'car': _loadout.get('body', None),
                'skin': _loadout.get('decal', None),
                'wheels': _loadout.get('wheels', None),
                'boost': _loadout.get('boost', _loadout.get('rocket_trail', None)),
                'antenna': _loadout.get('antenna', None),
                'topper': _loadout.get('topper', None),
                'engine_audio': _loadout.get('engine_audio', None),
                'trail': _loadout.get('trail', None),
                'goal_explosion': _loadout.get('goal_explosion', None),
                'banner': _loadout.get('banner', None)
            })
            # TODO: Support painted stuff (look in ClientLoadoutsOnline)
        logger.info('Loadout for %s: %s' % (self.name, self.loadout))

    def parse_data(self, _dict: dict):
        """
        ['ping', 'pos_x', 'pos_y', 'pos_z', 'rot_x', 'rot_y', 'rot_z', 'vel_x',
        'vel_y', 'vel_z', 'ang_vel_x', 'ang_vel_y', 'ang_vel_z', 'throttle',
        'steer', 'handbrake', 'ball_cam', 'dodge_active', 'double_jump_active',
        'jump_active', 'boost', 'boost_active']

        {'ang_vel_x': dtype('float64'),
         'ang_vel_y': dtype('float64'),
         'ang_vel_z': dtype('float64'),
         'ball_cam': dtype('O'),
         'boost': dtype('float64'),
         'boost_active': dtype('O'),
         'dodge_active': dtype('O'),
         'double_jump_active': dtype('O'),
         'handbrake': dtype('O'),
         'jump_active': dtype('O'),
         'ping': dtype('int64'),
         'pos_x': dtype('float64'),
         'pos_y': dtype('float64'),
         'pos_z': dtype('float64'),
         'rot_x': dtype('float64'),
         'rot_y': dtype('float64'),
         'rot_z': dtype('float64'),
         'steer': dtype('float64'),
         'throttle': dtype('float64'),
         'vel_x': dtype('float64'),
         'vel_y': dtype('float64'),
         'vel_z': dtype('float64')}

        :param _dict:
        :return:
        """
        self.data = pd.DataFrame.from_dict(_dict, orient='index')
        self.get_boost()

    def get_boost(self):
        # comparator has to be == for pandas
        if not ('boost_collect' in self.data.columns):
            self.data['boost_collect'] = np.nan
        boost_collection_frames = self.data.boost_collect[self.data.boost_collect == True].index.values
        for boost_collection_frame in boost_collection_frames:
            position = self.data.loc[boost_collection_frame, ['pos_x', 'pos_y', 'pos_z']]
            boost_type = get_if_full_boost_position(position)
            self.data.loc[boost_collection_frame, 'boost_collect'] = boost_type
